// ---------------------------------------------------------------------------
// Systems.java
// The translator of systems to their semantics as labeled transition systems.
// $Id: Systems.java,v 1.48 2020/10/22 15:25:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2020-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.semantics;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.*;
import java.util.function.*;

import riscal.util.*;
import riscal.semantics.Types.*;
import riscal.syntax.AST;
import riscal.Main;
import riscal.semantics.Translator.*;
import riscal.types.Environment.Symbol.*;
import riscal.types.*;

public class Systems
{
  // -------------------------------------------------------------------------
  //
  // a model checker for generic transition systems
  //
  // -------------------------------------------------------------------------

  // -------------------------------------------------------------------------
  // an abstract state
  // -------------------------------------------------------------------------
  public static interface State
  {
    // a representation of the run leading to the state using the given variable names
    public String getTrace(String[] vars);
  }

  // --------------------------------------------------------------------------
  // a (labeled) transition system with states from S
  // --------------------------------------------------------------------------
  public static interface LTS<S extends State>
  { 
    // the names of the state variables
    public String[] getVariables();

    // the initial states
    public Seq<S> getInitial();

    // the poststates of the given prestate
    public Seq<S> getNext(S state);
  }

  // provide access to currently used visitor
  private static Visitor<State> visitor = null;

  /**************************************************************************
   * Print trace leading to given context.
   * @param c the context.
   *************************************************************************/
  public static void printTrace(Context c)
  {
    if (visitor == null) return;
    visitor.printTrace(c);
  }

  // --------------------------------------------------------------------------
  // a class for generating all reachable states of a LTS with states from S
  // --------------------------------------------------------------------------
  public static class Visitor<S extends State>
  {
    private final LTS<S> lts;           // the labeled transition system
    private final Collection<S> result; // the result of the visit

    // maximum search depth and actual depth reached
    private final Long DEPTH = Main.getCheckDepth();
    private int depth;

    /**************************************************************************
     * Construct a fresh visitor for a system.
     * @param lts the system
     *************************************************************************/
    public Visitor(LTS<S> lts)
    {
      this.lts = lts;
      this.result = new HashSet<S>();
      this.depth = 0;
    }

    // the time when last output was produced
    private long time;
    
    /**************************************************************************
     * Visit all reachable system states.
     * @return the collection of reachable states.
     *************************************************************************/
    @SuppressWarnings("unchecked")
    public Collection<S> visit()
    {
      visitor = (Visitor<State>)this;
      time = System.currentTimeMillis();
      visit(lts.getInitial());
      visitor = null;
      return result;
    }

    /**************************************************************************
     * Return the depth of the execution tree.
     * @return the depth of the tree, i.e., the maximum number of transitions.
     *************************************************************************/
    public int getDepth()
    {
      return depth;
    }

    /**************************************************************************
     * Visit all states reachable from the given states
     * @param states a stream of states
     *************************************************************************/
    private void visit(Seq<S> states)
    {
      while (true)
      {
        Seq.Next<S> next;
        try
        {
          next = states.get();
        }
        catch(StateError e)
        {
          throw new RuntimeError(e.tree, e.getMessage() +
              "\nThe system run leading to this error:\n" + e.state.getTrace(lts.getVariables()));
        }
        if (next == null) return;
        visit(next.head, 0);
        states = next.tail;
      }
    }

    /**************************************************************************
     * Visit all states reachable from the given state.
     * @param state the given state.
     * @param depth the current search depth.
     *************************************************************************/
    private void visit(S state, int depth)
    {
      if (result.contains(state)) return;
      result.add(state);
      if (DEPTH != null && depth+1 > DEPTH) return;
      if (depth+1 > this.depth) this.depth = depth+1;
      long time0 = System.currentTimeMillis();
      if (time0 - time >= 2000)
      {
        PrintWriter writer = Main.getOutput();
        writer.println(result.size() + " system states visited...");
        time = time0;
      }
      Seq<S> states = lts.getNext(state);
      while (true)
      {
        Seq.Next<S> next;
        try
        {
          next = states.get();
        }
        catch(StateError e)
        {
          throw new RuntimeError(e.tree, e.getMessage() +
              "\nThe system run leading to this error:\n" + e.state.getTrace(lts.getVariables()));
        }
        catch(TraceError e)
        {
          throw new RuntimeError(e.tree, e.getMessage() +
              "\nThe system run leading to this error:\n" + state.getTrace(lts.getVariables()) + e.trace + "->...");
        }
        catch(RuntimeError e)
        {
          throw new RuntimeError(e.tree, e.getMessage() +
              "\nThe system run leading to this error:\n" + state.getTrace(lts.getVariables()) + "->...");
        }
        catch(Exception e)
        {
          StringWriter w = new StringWriter();
          e.printStackTrace(new PrintWriter(w));
          throw new RuntimeException(w.toString() +
              "The system run leading to this error:\n" + state.getTrace(lts.getVariables()) + "->...");
        }
        if (next == null) break;
        visit(next.head, depth+1);
        states = next.tail;
      }
    }    
    
    /**************************************************************************
     * Print trace leading to given context.
     * @param c the context.
     *************************************************************************/
    public void printTrace(Context c)
    {
      StateClass parent = c.getParent();
      Label label = c.getLabel();
      String[] vars = lts.getVariables();
      StateClass state = new StateClass(c, vars.length, parent, label);
      Main.println(state.getTrace(vars));
    }
  }
  
  // an error that gives us a more specific state
  private static class StateError extends RuntimeError
  {
    private static final long serialVersionUID = 05032020L;
    private final State state;
    public StateError(AST tree, String message, State state)
    {
      super(tree, message);
      this.state = state;
    }
  }

  // an error that gives us some extra trace
  private static class TraceError extends RuntimeError
  {
    private static final long serialVersionUID = 05032020L;
    private final String trace;
    public TraceError(AST tree, String message, String trace)
    {
      super(tree, message);
      this.trace = trace;
    }
  }

  // -------------------------------------------------------------------------
  //
  // the implementation of concrete transition systems
  //
  // -------------------------------------------------------------------------

  // -------------------------------------------------------------------------
  // a transition label
  // -------------------------------------------------------------------------
  public static class Label
  {
    public final ActionSymbol action; // the action
    public final Value[] args;        // the values of the action arguments
    public final int instance;        // the instance number
    
    public Label(ActionSymbol action, Value[] args, int instance)
    {
      this.action = action;
      this.args = args;
      this.instance = instance;
    }
    
    public Label(ActionSymbol action, Value[] args)
    {
      this (action, args, -1);
    }

    public String toString()
    {
      ComponentSymbol component = action.getComponent();
      return (component == null ? "" : component.ident.string + 
          (instance == -1 ? "" : "[" + instance + "]") + ".") +
          action.ident.string + "(" + AST.toString(args, ",") + ")";
    }
  }

  // -------------------------------------------------------------------------
  // a concrete state
  // -------------------------------------------------------------------------
  public static class StateClass implements State
  {
    // the context from which this state emerged
    public final Context context;

    // the values of the state variables
    public final Value[] values;  

    // the predecessor state (null, if state is initial)
    public StateClass parent;

    // the label of the transition leading to this state (null, if state is initial)
    public Label label; 

    /**************************************************************************
     * Construct an initial state from a context.
     * @param c the context.
     * @param n the number of state variables.
     *************************************************************************/
    public StateClass(Context c, int n)
    {
      this(c, n, null, null);
    }

    /**************************************************************************
     * Construct a state from a context.
     * @param c the context.
     * @param n the number of state variables.
     * @param parent the predecessor state (null if this state is initial).
     * @param label the label of the transition leading from the predecessor
     *        state to this state (null if this state is initial).
     *************************************************************************/
    public StateClass(Context c, int n, StateClass parent, Label label)
    {
      this.context = c;
      Value[] values = new Value[n];
      for (int i = 0; i < n; i++) values[i] = c.get(i);
      this.values = values;
      this.parent = parent;
      this.label = label;
    }

    /**************************************************************************
     * Functions to support quick equality tests/hash table lookups
     * (based only on the values of the state variables).
     *************************************************************************/
    public boolean equals(Object o) 
    { 
      if (!(o instanceof StateClass)) return false;
      StateClass s = (StateClass)o;
      int n = values.length;
      if (n != s.values.length) return false;
      for (int i=0; i<n; i++)
      {
        if (values[i] == null)
        {
          if (s.values[i] != null) return false;
          continue;
        }
        if (!values[i].equals(s.values[i])) return false;
      }
      return true;
    }
    public int hashCode() 
    { 
      return Objects.hash(Arrays.hashCode(values)); 
    }

    /**************************************************************************
     * The string representation of the state using given variable names.
     * @param vars the names of the state variables.
     *************************************************************************/
    public String toString(String[] vars)
    {
      StringBuffer result = new StringBuffer();
      result.append("[");
      int n = vars.length;
      for (int i = 0; i < n; i++)
      {
        result.append(vars[i] + ":" + values[i]);
        if (i < n-1) result.append(",");
      }
      result.append("]");
      return result.toString();
    }

    /**************************************************************************
     * Get trace leading to this state.
     * @param vars the names of the state variables.
     * @return the trace
     *************************************************************************/
    public String getTrace(String[] vars)
    {
      StringBuffer result = new StringBuffer();
      getTrace(vars, result);
      return result.toString();
    }

    /**************************************************************************
     * Get trace leading to this state.
     * @param vars the names of the state variables.
     * @param result the buffer to which the trace is appended.
     *************************************************************************/
    private int getTrace(String[] vars, StringBuffer result)
    {
      int depth;   
      if (parent == null)
        depth = 0;
      else
      {
        depth = 1+parent.getTrace(vars, result);
        result.append("->" + label + "->\n");
      }
      result.append("  " + depth + ":" + toString(vars));
      return depth;
    }

    /**************************************************************************
     * Get root of state tree.
     * @return the ancestor of this state that has no parent.
     *************************************************************************/
    public StateClass getRoot()
    {
      StateClass root = this;
      while (true)
      {
        StateClass parent = root.parent;
        if (parent == null) return root;
        root = parent;
      }     
    }
  }

  // -------------------------------------------------------------------------
  //
  // the translation of shared system declarations to transition systems
  //
  // -------------------------------------------------------------------------

  // the system semantics is a labeled transition system
  public interface SysSem extends LTS<StateClass>{ }

  // the action semantics is a function from states to sequences of states
  public interface ActSem extends Function<StateClass,Seq<StateClass>> { }

  // the overall translator
  private Translator translator;

  /***************************************************************************
   * Create systems translator.
   * @param translator the overall translator.
   **************************************************************************/
  public Systems(Translator translator)
  {
    this.translator = translator;
  }

  /****************************************************************************
   * Translate a shared system to a labeled transition system.
   * @param system the system (to be annotated with the resulting LTS).
   * @param exps the initialization expressions (each of which may be null).
   * @param specs the system specifications.
   * @param iaction the initialization action (may be null).
   * @param actions the system actions.
   ***************************************************************************/
  public void SharedSystem(SystemSymbol system, 
    ExpSem[]exps, SysSpecSem[] specs, ActionSymbol iaction, ActionSymbol[] actions)
  {     
    // in the following we populate the system symbol with some extra information

    // the variable names
    ValueSymbol[] vars = system.getVars();
    String[] names = Arrays.stream(vars)
        .map((ValueSymbol v)->v.ident.string).toArray(String[]::new);
    system.setNames(names);

    // the initial states
    int n = vars.length;
    int size = 0;
    for (int i = 0; i < n; i++)
    {
      if (exps[i] == null) 
        exps[i] = (ExpSem.Single)(Context c)->null; // new Value.Unit();
      else  
      {
        int slots = vars[i].getSlots();
        if (slots > size) size = slots;
      }
    }
    Seq<Context> contexts0 = Commands.apply(exps, 
        (Value[] vs, Context c)-> newContext(system, vs)).toMultiple().apply(new Context(size));
    Seq<StateClass> istates0 = contexts0.apply((Context c)->new StateClass(c,n));
    ActSem iaction0 = iaction != null ? iaction.getValue() : (StateClass state)->
    Seq.cons(stateFun(system).apply(newContext(system, state.values)), Seq.empty());
    Seq<StateClass> istates1 = istates0.applyJoin(iaction0);

    // the next states
    Function<StateClass,Seq<StateClass>> next = (StateClass s) -> Seq.empty();
    int m = actions.length;
    for (int i = m-1; i >= 0; i--)
    {
      ActSem action = actions[i].getValue();
      final Function<StateClass,Seq<StateClass>> next0 = next;
      next = (StateClass s)->Seq.append(action.apply(s), next0.apply(s));
    }
    final Function<StateClass,Seq<StateClass>> next0 = next;

    // the resulting system
    system.setValue(new SysSem() {
      public String[] getVariables () 
      { return names; }
      public Seq<StateClass> getInitial() 
      { return istates1; }
      public Seq<StateClass> getNext(StateClass state)
      { return next0.apply(state); }
    });
  }

  // -------------------------------------------------------------------------
  //
  // Actions
  //
  // -------------------------------------------------------------------------

  public void CommandAction(ActionSymbol action, ExpSem exp, ComSem command)
  {
    Predicate<Context> filterPredicate = filterPredicate(exp);
    Function<Context,Context> cloneFun = (Context c)-> new Context(c);
    Function<Context,StateClass> stateFun = stateFun(action.system);
    Function<Context,Seq<StateClass>> resultFun = applySem(command, stateFun);
    action.setValue(new ActSem() {
      public Seq<StateClass> apply(StateClass state)
      {
        // clone context before application of command to avoid false sharing
        return initFun(action,state).filter(filterPredicate)
            .apply(cloneFun).applyJoin(resultFun);
      }
    });
  }

  public void FunctionAction(ActionSymbol action, ExpSem exp, ExpSem fexp)
  {
    Predicate<Context> filterPredicate = filterPredicate(exp);
    Function<Context,Seq<Context>> valueFun = applySem(fexp, valueFun(action.system));
    Function<Context,StateClass> stateFun = stateFun(action.system);
    action.setValue(new ActSem() {
      public Seq<StateClass> apply(StateClass state)
      {
        return initFun(action,state).filter(filterPredicate)
            .applyJoin(valueFun).apply(stateFun);
      }
    });
  }

  public void PredicateAction(ActionSymbol action, ExpSem exp, ExpSem pexp)
  {
    SystemSymbol system = action.system;
    Predicate<Context> filterPredicate = filterPredicate(exp);
    Function<Context,Seq<Context>> generateFun = generateFun(action.system.getNewVars());
    Predicate<Context> filterPredicate2 = filterPredicate(pexp);
    Function<Context,Context> assignFun = assignFun(system.getVars(), system.getNewVars());
    Function<Context,StateClass> stateFun = stateFun(action.system);
    action.setValue(new ActSem() {
      public Seq<StateClass> apply(StateClass state)
      {
        return initFun(action,state).filter(filterPredicate)
            .applyJoin(generateFun).filter(filterPredicate2)
            .apply(assignFun).apply(stateFun);
      }
    });
  }

  // -------------------------------------------------------------------------
  //
  // Basic functions from which semantics is composed
  //
  // -------------------------------------------------------------------------

  /***************************************************************************
   * Construct all possible initial contexts of action.
   * @param action the action.
   * @param state the state to which the action is applied.
   * @return the initial contexts.
   **************************************************************************/
  private Seq<Context> initFun(ActionSymbol action, StateClass state)
  {
    SystemSymbol system = action.system;
    List<ContextFunction> variants = system.getVariants();
    List<AST> cvariants = system.getVariantDeclarations();
    int pn = action.params.length;
    Seq<Value> args = Values.getValueSeq(new AST.Type.TupleType(action.types));
    Seq<Context> contexts = args.apply((Value v)-> {
      Context c = newContext(system, state.values);
      Value.Array v0 = (Value.Array)v;
      for (int i = 0; i < pn; i++)
        c.set(action.params[i].getSlot(), v0.get(i), action.params[i].ident.string);
      if (!action.ident.string.equals("_init"))
      {
        c.setSystemMeasures(translator.initMeasures(c, variants, cvariants));
        c.setParent(state);
        c.setLabel(new Label(action, v0.get()));
      }
      return c;
    });
    return contexts;
  }

  /***************************************************************************
   * Construct context predicate from expression.
   * @param pexp the semantics of the predicate expression.
   * @return the predicate on contexts.
   *************************************************************************/
  private Predicate<Context> filterPredicate(ExpSem pexp)
  {
    if (pexp == null)
    {
      return (Context c)->true;
    }
    else if (pexp instanceof ExpSem.Single)
    {
      ExpSem.Single pexp0 = (ExpSem.Single)pexp;
      return (Context c)->
      {
        try
        {
          Value.Bool v = (Value.Bool)pexp0.apply(c);
          return v.getValue();
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      };
    }
    else
    {
      ExpSem.Multiple pexp0 = (ExpSem.Multiple)pexp;
      return (Context c)-> 
      {
        try
        {
          Seq<Value> vs = pexp0.apply(c);
          while (true)
          {
            Seq.Next<Value> next = vs.get();
            if (next == null) return false;
            Value.Bool v0 = (Value.Bool)next.head;
            if (v0.getValue()) return true;
            vs = next.tail;
          }
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      };
    }
  }

  /***************************************************************************
   * Get function that maps contexts to states.
   * @param system the system
   * @return the mapping function.
   *************************************************************************/
  private Function<Context,StateClass> stateFun(SystemSymbol system)
  {
    List<ContextRelation> invariants = system.getInvariants();
    List<ContextFunction> variants = system.getVariants();
    List<AST> cinvariants = system.getInvariantDeclarations();
    List<AST> cvariants = system.getVariantDeclarations();
    int n = system.getVars().length;
    return (Context c)->
    {
      StateClass parent = c.getParent();
      Label label = c.getLabel();
      Context c0 = parent == null ? c : parent.getRoot().context;
      StateClass result = new StateClass(c, n, parent, label);
      try
      {
        translator.checkInvariant(invariants, c, c0, cinvariants);
        List<List<Value[]>> measures = c.getSystemMeasures();
        if (measures != null)
          translator.checkMeasures(measures, c, variants, cvariants);
        Main.checkStopped();
      }
      catch(RuntimeError e)
      {
        throw new StateError(e.tree, e.getMessage(), result);
      }
      return result;
    };   
  }

  /***************************************************************************
   * Get function that maps contexts to functions from values to contexts
   * where the result context has its extra information from the original
   * context and its values from the given values.
   * @param system the system
   * @return the mapping function.
   *************************************************************************/
  private Function<Context,Function<Value,Context>> 
  valueFun(SystemSymbol system)
  {
    int n = system.getVars().length;
    return (Context c)->(Value v)->
    {
      Context c0 = n == 1 ? newContext(system, new Value[] {v}) :
        newContext(system, ((Value.Array)v).get()); 
      c0.setLabel(c.getLabel());
      c0.setParent(c.getParent());
      c0.setSystemMeasures(c.getSystemMeasures());
      return c0;
    };
  }

  /***************************************************************************
   * Generate all possible successor contexts of action.
   * @param variables the variables to be set to all possible values.
   * @return a function from the initial context to the result contexts.
   **************************************************************************/
  private Function<Context,Seq<Context>> generateFun(ValueSymbol[] vars)
  {
    return (Context c)->
    {
      int n = vars.length;
      AST.Type[] types = new AST.Type[n];
      for (int i = 0; i < n; i++)
        types[i] = vars[i].type;
      Seq<Value> args = Values.getValueSeq(new AST.Type.TupleType(types));
      Seq<Context> contexts = args.apply((Value v)-> {
        Context c0 = new Context(c);
        Value.Array v0 = (Value.Array)v;
        for (int i = 0; i < n; i++)
          c0.set(vars[i].getSlot(), v0.get(i), vars[i].ident.string);
        return c0;
      });
      return contexts;
    };
  }

  /***************************************************************************
   * Assign to prestate variables the values of the corresponding poststate variables.
   * @param ovars the prestate ("old") variables
   * @param nvars the poststate ("new") variables
   * @return a function that takes a context and writes into the locations
   * of the state variables the values from their poststate locations. 
   *************************************************************************/
  private Function<Context,Context> assignFun(ValueSymbol[] ovars, ValueSymbol[] nvars)
  {
    // we update the context itself rather than generating a clone
    // this is okay if the contexts were generated from scratch by generateFun()
    return (Context c)->
    {
      int n = ovars.length;
      for (int i = 0; i < n; i++)
      {
        c.set(ovars[i].getSlot(), c.get(nvars[i].getSlot()), ovars[i].ident.string);
      }
      return c;
    };
  }

  // -------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // -------------------------------------------------------------------------

  /***************************************************************************
   * Create new context with initialized state variables.
   * @param system the system to which the context belongs.
   * @param values the initial values of the state variables.
   * @return the new context.
   **************************************************************************/
  private static Context newContext(SystemSymbol system, Value[] values)
  {
    Context c = new Context(system.getSlots());
    String[] names = system.getNames();
    int n = names.length;
    for (int i = 0; i < n; i++) c.set(i, values[i], names[i]);
    return c;
  }

  /***************************************************************************
   * Apply a function to the semantics of a command. 
   * @param com the command.
   * @param fun the function.
   * @return a function that maps a context c to all fun(com(c))
   **************************************************************************/
  private static <T> Function<Context,Seq<T>> 
  applySem(ComSem com, Function<Context,T> fun)
  {
    if (com instanceof ComSem.Single)
    {
      ComSem.Single com0 = (ComSem.Single)com;
      return (Context c)->
      {
        try
        {
          return Seq.cons(fun.apply(com0.apply(c)),Seq.empty());
        }
        catch(RuntimeError e)
        {
          if (e instanceof StateError) throw e;
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      };
    }
    else
    {
      ComSem.Multiple com0 = (ComSem.Multiple)com;
      return (Context c)-> 
      {
        try
        {
          return com0.apply(c).apply(fun);
        }
        catch(RuntimeError e)
        {
          if (e instanceof StateError) throw e;
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      };
    }
  }

  /***************************************************************************
   * Apply a function to the semantics of an expression.
   * @param exp the expression.
   * @param fun the function.
   * @return a function that maps a context c to all fun(c)(exp(c))
   **************************************************************************/
  private static <T> Function<Context,Seq<T>> 
  applySem(ExpSem exp, Function<Context,Function<Value,T>> fun)
  {
    if (exp instanceof ExpSem.Single)
    {
      ExpSem.Single exp0 = (ExpSem.Single)exp;
      return (Context c)->
      {
        try
        {
          return Seq.cons(fun.apply(c).apply(exp0.apply(c)),Seq.empty());
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      };
    }
    else
    {
      ExpSem.Multiple exp0 = (ExpSem.Multiple)exp;
      return (Context c)-> 
      {
        try
        {
          return exp0.apply(c).apply(fun.apply(c));
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      };
    }
  }

  /****************************************************************************
   * Get all invariants from specifications.
   * @param spec the specifications.
   * @return the invariants.
   ***************************************************************************/
  public static List<AST> getInvariants(AST.SystemSpec[] spec)
  {
    List<AST> result = new ArrayList<AST>();
    for (AST.SystemSpec s : spec)
    {
      if (s instanceof AST.SystemSpec.InvariantSystemSpec) 
        result.add(s);
    }
    return result;
  }

  /****************************************************************************
   * Get all invariants from specifications.
   * @param spec the specifications.
   * @return the invariants.
   ***************************************************************************/
  public static List<ContextRelation> getInvariants(SysSpecSem[] spec)
  {
    List<ContextRelation> result = new ArrayList<ContextRelation>();
    for (SysSpecSem s : spec)
    {
      if (s instanceof SysSpecSem.Invariant) 
        result.add((ContextRelation)s);
    }
    return result;
  }

  /****************************************************************************
   * Get all variants from specifications.
   * @param spec the specifications.
   * @return the variants.
   ***************************************************************************/
  public static List<AST> getVariants(AST.SystemSpec[] spec)
  {
    List<AST> result = new ArrayList<AST>();
    for (AST.SystemSpec s : spec)
    {
      if (s instanceof AST.SystemSpec.DecreasesSystemSpec) 
        result.add(s);
    }
    return result;
  }

  /****************************************************************************
   * Get all variants from specifications.
   * @param spec the specifications.
   * @return the variants.
   ***************************************************************************/
  public static List<ContextFunction> getVariants(SysSpecSem[] spec)
  {
    List<ContextFunction> result = new ArrayList<ContextFunction>();
    for (SysSpecSem s : spec)
    {
      if (s instanceof SysSpecSem.Variant) 
        result.add((ContextFunction)s);
    }
    return result;
  }

  // -------------------------------------------------------------------------
  //
  // the translation of distributed system declarations to transition systems
  //
  // -------------------------------------------------------------------------

  /****************************************************************************
   * Translate a distributed system to a labeled transition system.
   * @param system the system (to be annotated with the resulting LTS).
   * @param specs the system specifications.
   * @param decl the original system declaration.
   ***************************************************************************/
  public void DistributedSystem(SystemSymbol system, 
    SysSpecSem[] specs, AST.Declaration.DistributedSystem decl)
  { 
    // in the following we populate the system symbol with some extra information
    ComponentSymbol[] components = system.getComponents();
    
    // the variable names
    ValueSymbol[] vars = system.getVars();
    String[] names = Arrays.stream(vars)
        .map((ValueSymbol v)->v.ident.string).toArray(String[]::new);
    system.setNames(names);
    
    // the initial contexts arising from initializing the components in turn
    Function<Context,Seq<Context>> init = (Context c) -> Seq.cons(c,Seq.empty());
    int n = components.length;
    for (int i = 0; i < n; i++)
    {
      Function<Context,Seq<Context>> component = components[i].getValue().getInitial();
      final Function<Context,Seq<Context>> init0 = init;
      init = (Context c)->init0.apply(c).applyJoin(component);
    }
    final Function<Context,Seq<Context>> init0 = init;

    // the next contexts
    Function<Context,Seq<Context>> next = (Context c) -> Seq.empty();
    for (int i = n-1; i >= 0; i--)
    {
      Function<Context,Seq<Context>> component = components[i].getValue().getNext();
      final Function<Context,Seq<Context>> next0 = next;
      next = (Context c)->Seq.append(component.apply(c), next0.apply(c));
    }
    final Function<Context,Seq<Context>> next0 = next;

    // the corresponding system entities
    final Seq<StateClass> inits = 
        init0.apply(newContextDist(system)).apply(stateFun(system));
    final Function<StateClass,Seq<StateClass>> nexts = (StateClass s)->
      next0.apply(newContextDist(system,s)).apply(stateFun(system));

    // the resulting system
    system.setValue(new SysSem() {
      public String[] getVariables () 
      { return names; }
      public Seq<StateClass> getInitial() 
      { return inits; }
      public Seq<StateClass> getNext(StateClass state)
      { return nexts.apply(state); }
    });
  }

  /***************************************************************************
   * Create context for system with empty action buffers.
   * @param system the system.
   * @return the context.
   **************************************************************************/
  private static Context newContextDist(SystemSymbol system)
  {
    Context c = new Context(system.getSlots());
    ComponentSymbol[] components = system.getComponents();
    int sn = components.length;
    for (int i = 0; i < sn; i++)
    {
      ComponentSymbol component = components[i];
      int m = component.getVars().length;
      List<ActionSymbol> actions = component.getActions();
      int n = actions.size();
      Value.Array cstate = new Value.Array(m+2*n);
      for (int j = 0; j < n; j++)
      {
        int bn = actions.get(j).bsize;
        Value.Array buffer = new Value.Array(bn);
        cstate.set(m+2*j, buffer);
        cstate.set(m+2*j+1, new Value.Int(0));
      }
      if (component.instances != 0)
      {
        int cn = component.instances;
        Value.Array cstate0 = new Value.Array(cn);
        for (int j = 0; j < cn; j++)
          cstate0.set(j, new Value.Array(cstate));
        cstate = cstate0;
      }
      c.set(i, cstate, component.ident.string);
    }    
    return c;
  }

  /***************************************************************************
   * Create new context from given state.
   * @param system the system to which the context belongs.
   * @param values the state.
   * @return the new context (linked to state).
   **************************************************************************/
  private static Context newContextDist(SystemSymbol system, StateClass state)
  {
    Context c = newContext(system, state.values);
    c.setParent(state);
    return c;
  }
  
  // --------------------------------------------------------------------------
  // the semantics of a component
  // --------------------------------------------------------------------------
  public static interface CompSem
  { 
    // the contexts produced by initializing this component in the given context
    public Function<Context,Seq<Context>> getInitial();

    // the contexts produced by a step of this component in the given context
    public Function<Context,Seq<Context>> getNext();
  }

  // the  component action semantics is a function from contexts to sequences of contexts
  public interface ActSemComp extends Function<Context,Seq<Context>> { }

  /****************************************************************************
   * Translate a component to its semantics (with which it is annotated).
   * @param component the component.
   * @param exps the initialization expressions (each of which may be null).
   * @param iaction the initialization action (may be null).
   * @param actions the component actions.
   ***************************************************************************/
  public void Component(ComponentSymbol component, 
    ExpSem[]exps, ActionSymbol iaction, ActionSymbol[] actions)
  { 
    // the variable names
    ValueSymbol[] vars = component.getVars();
    String[] names = Arrays.stream(vars)
        .map((ValueSymbol v)->v.ident.string).toArray(String[]::new);
    component.setNames(names);

    // the initial contexts
    int n = exps.length;
    for (int i = 0; i < n; i++)
    {
      if (exps[i] != null) continue;
      exps[i] = (ExpSem.Single)(Context c)->null; // new Value.Unit();
    }
    Function<Context,Seq<Context>> init1 = Commands.apply(exps, 
        (Value[] vs, Context c)-> newContext(component, c, vs)).toMultiple();
    Function<Context,Context> transfer = transferFun(component);
    Function<Context,Seq<Context>> init2 = 
        iaction == null ? initFun(component) : initFun(iaction);
    Function<Context,Seq<Context>> init0 =
        (Context c) -> init1.apply(c).applyJoin(init2);
    
    // the next contexts
    Function<Context,Seq<Context>> next = (Context c) -> Seq.empty();
    int m = actions.length;
    for (int i = m-1; i >= 0; i--)
    {
      ActionSymbol action = actions[i];
      ActSemComp asem = action.getValueComp();
      Function<Context,Seq<Context>> init = nextFun(action);
      final Function<Context,Seq<Context>> next0 = next;
      next = (Context c)->Seq.append(init.apply(c)
          .applyJoin(asem).filter((Context c0)-> !c0.isNull()).apply(transfer), 
          next0.apply(c));
    }
    final Function<Context,Seq<Context>> next0 = next;

    // the component semantics
    component.setValue(new CompSem() {
      public Function<Context,Seq<Context>> getInitial() 
      { return init0; }
      public Function<Context,Seq<Context>> getNext()
      { return next0; }
    });
  }

  //-------------------------------------------------------------------------
  //
  // Actions (component semantics)
  //
  // -------------------------------------------------------------------------

  public void CommandActionComponent(ActionSymbol action, ExpSem exp, ComSem command)
  {
    Predicate<Context> filterPredicate = filterPredicate(exp);
    action.setValueComp(new ActSemComp() {
      public Seq<Context> apply(Context c)
      {
        try
        {
          if (!filterPredicate.test(c)) return Seq.empty();
          c = new Context(c);
          return command.toMultiple().apply(c);
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      }
    });
  }

  public void FunctionActionComponent(ActionSymbol action, ExpSem exp, ExpSem fexp)
  {
    Predicate<Context> filterPredicate = filterPredicate(exp);
    Function<Context,Function<Value,Context>> valueFun = valueFun(action.getComponent());
    action.setValueComp(new ActSemComp() {
      public Seq<Context> apply(Context c)
      {
        try
        {
          if (!filterPredicate.test(c)) return Seq.empty();
          return fexp.toMultiple().apply(c).apply(valueFun.apply(c));
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      }
    });
  }

  public void PredicateActionComponent(ActionSymbol action, ExpSem exp, ExpSem pexp)
  {
    ComponentSymbol component = action.getComponent();
    Predicate<Context> filterPredicate = filterPredicate(exp);
    Function<Context,Seq<Context>> generateFun = generateFun(component.getNewVars());
    Predicate<Context> filterPredicate2 = filterPredicate(pexp);
    Function<Context,Context> assignFun = assignFun(component.getVars(), component.getNewVars());
    action.setValueComp(new ActSemComp() {
      public Seq<Context> apply(Context c)
      {
        try
        {
          if (!filterPredicate.test(c)) return Seq.empty();
          return generateFun.apply(c).filter(filterPredicate2).apply(assignFun);
        }
        catch(RuntimeError e)
        {
          throw new TraceError(e.tree, e.getMessage(), "->" + c.getLabel());
        }
      }
    });
  }

  //-------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // -------------------------------------------------------------------------

  /***************************************************************************
   * Return the context slot offset for the local state of the component.
   * @param component the component.
   * @return the index of the first state variable of the component.
   **************************************************************************/
  private static int componentOffset(ComponentSymbol component)
  {
    // the persistent part of the context has two entries for every component
    return 2*component.system.getComponents().length;
  }

  /***************************************************************************
   * Create local component context from global system context 
   * by initializing the state variables of the component.
   * @param component the component.
   * @param c the global system context.
   * @param values the values of the state variables.
   * @return the local component context (an updated clone of the system context).
   **************************************************************************/
  private static Context newContext(ComponentSymbol component, Context c, Value[] values)
  {
    int offset = componentOffset(component);
    String[] names = component.getNames();
    int n = values.length;
    Context c0 = new Context(c);
    // offset 1 to consider "this"
    for (int i = 0; i < n; i++) c0.set(1+offset+i, values[i], names[i]);  
    return c0;
  }

  /***************************************************************************
   * Construct initial contexts for component without initialization action.
   * @param component the component.
   * @return a function from the global system context to the initial
   * local contexts after execution of the action.
   **************************************************************************/
  private Function<Context,Seq<Context>> initFun(ComponentSymbol component)
  {
    boolean replicated = component.instances != 0;
    int ci = replicated ? component.instances : 1;
    int offset = componentOffset(component);
    Function<Context,Context> transfer = transferFun(component);
    return (Context c) -> 
    {
      // initialize all instances of the component in turn
      Seq<Context> cs = Seq.cons(new Context(c), Seq.empty());
      for (int i = 0; i < ci; i++)
      {
        int thisNumber = i;
        Function<Context,Context> initFun = (Context c0) -> {
          Context c1 = new Context(c0);
          c1.set(offset, new Value.Int(thisNumber), TypeChecker.thisName);
          c1.setNull(false);
          return transfer.apply(c1);
        };
        cs = cs.apply(initFun);
      }
      return cs;
    };
  }
  
  /***************************************************************************
   * Construct initial contexts from execution of initialization action.
   * @param iaction a component's initialization action.
   * @return a function from the global system context to the initial
   * local contexts after execution of the action.
   **************************************************************************/
  private Function<Context,Seq<Context>> initFun(ActionSymbol iaction)
  {
    ComponentSymbol component = iaction.getComponent();
    boolean replicated = component.instances != 0;
    int ci = replicated ? component.instances : 1;
    String[] names = component.getNames();
    int n = names.length;
    int offset = componentOffset(component);
    int pn = iaction.params.length;
    Function<Context,Context> transfer = transferFun(component);
    Seq<Value> args = Values.getValueSeq(new AST.Type.TupleType(iaction.types));
    ActSemComp sem = iaction.getValueComp();
    return (Context c) -> 
    {
      // initialize all instances of the component in turn
      Seq<Context> cs = Seq.cons(new Context(c), Seq.empty());
      for (int i = 0; i < ci; i++)
      {
        int thisNumber = i;
        Function<Context,Seq<Context>> initFun = (Context c0) -> args.applyJoin((Value v)-> {
          Context c1 = new Context(c0);
          // initialize "this" with instance number
          c1.set(offset, new Value.Int(thisNumber), TypeChecker.thisName);
          // initialize parameters with lazily generated sequence value
          Value.Array v0 = (Value.Array)v;
          for (int j = 0; j < pn; j++)
            c1.set(1+offset+2*n+j, v0.get(j), iaction.params[j].ident.string);
          c1.setNull(false);
          return sem.apply(c1).filter((Context c2)-> !c2.isNull()).apply(transfer);
        });
        cs = cs.applyJoin(initFun);
      }
      return cs;
    };
  }
  
  /***************************************************************************
   * Apply function to every integer i with 0 <= i < n
   * @param n the upper bound of the arguments.
   * @param f the function.
   * @return the sequence of result values.
   **************************************************************************/
  private static <T> Seq<T> apply(int n, Function<Integer,T> f)
  {
    Seq<Integer> args = Seq.empty();
    for (int i = n-1; i >= 0; i--) args = Seq.cons(i, args);
    return args.apply(f);
  }
  
  /***************************************************************************
   * Construct all possible initial contexts of a component action.
   * @param action the action.
   * @return a function from the global system context to the initial
   * local contexts in which the action may start execution.
   **************************************************************************/
  private Function<Context,Seq<Context>> nextFun(ActionSymbol action)
  {
    ComponentSymbol component = action.getComponent();
    SystemSymbol system = component.system;
    List<ContextFunction> variants = system.getVariants();
    List<AST> cvariants = system.getVariantDeclarations();
    boolean replicated = component.instances != 0;
    int ci = replicated ?  component.instances : 1;
    String[] names = component.getNames();
    int n = names.length;
    int offset = componentOffset(component);
    int pn = action.params.length;
    int ai = action.getNumber();
    int bsize = action.bsize;
    int cn = component.number;
    String name = component.ident.string;
    // lazily create the contexts for all instances of the component
    return (Context c) -> apply(ci, (Integer i)->
    {
      Value.Array ostate = (Value.Array)c.get(cn);
      Value.Array cstate = replicated ? (Value.Array)ostate.get(i) : ostate; 
      Value.Array buffer = (Value.Array)cstate.get(n+2*ai);
      Value.Int number = (Value.Int)cstate.get(n+2*ai+1);
      int number0 = number.getValue();
      // if there is no message in the buffer, the action is not enabled
      if (number0 == 0) return null;
      Context c0 = new Context(c);
      // get old values of measures (*before* setting local component context)
      c0.setSystemMeasures(translator.initMeasures(c0, variants, cvariants));
      // initialize "this" with instance number
      c0.set(offset, new Value.Int(i), TypeChecker.thisName);
      // initialize local component context from global system context      
      for (int j = 0; j < n; j++)
        c0.set(1+offset+j, cstate.get(j), names[j]);
      // extract message from buffer
      Value.Array v = (Value.Array)buffer.get(0);
      // shift other messages (circular buffer implementation would be also possible)
      Value.Array buffer0 = new Value.Array(bsize);
      for (int k = 0; k < bsize-1; k++) buffer0.set(k, buffer.get(k+1));
      // buffer0.set(bsize-1, new Value.Unit());
      // write updated buffer and number back into context
      Value.Array cstate0 = new Value.Array(cstate);
      cstate0.set(n+2*ai, buffer0);
      cstate0.set(n+2*ai+1, new Value.Int(number0-1));
      if (replicated)
      {
        Value.Array cstate1 = new Value.Array(ostate);
        cstate1.set(i, cstate0);
        cstate0 = cstate1;
      }
      c0.set(cn, cstate0, name);
      // initialize parameters with received message
      for (int j = 0; j < pn; j++)
        c0.set(1+offset+2*n+j, v.get(j), action.params[j].ident.string);
      // set label of transition
      int instance = replicated ? i : -1;
      c0.setLabel(new Label(action, v.get(), instance));
      c0.setNull(false);
      return c0;
    }).filter((Context c0)->c0 != null);
  }
  
  /***************************************************************************
   * Get function that maps contexts to functions from values to contexts
   * where the result context has its extra information from the original
   * context and its values from the given values.
   * @param component the component.
   * @return the mapping function.
   *************************************************************************/
  private Function<Context,Function<Value,Context>> 
  valueFun(ComponentSymbol component)
  {
    int n = component.getVars().length;
    return (Context c)->(Value v)->
    {
      Context c0 = n == 1 ? newContext(component, c, new Value[] {v}) :
        newContext(component, c, ((Value.Array)v).get()); 
      c0.setLabel(c.getLabel());
      c0.setParent(c.getParent());
      c0.setSystemMeasures(c.getSystemMeasures());
      return c0;
    };
  }

  /***************************************************************************
   * Get function that transfers state of component to global state.
   * @param component the component.
   * @return the transfer function.
   *************************************************************************/
  private Function<Context,Context> transferFun(ComponentSymbol component)
  {
    String cname = component.ident.string;
    String[] names = component.getNames();
    int n = names.length;
    int an = component.getActions().size();
    int cn = component.number;
    boolean replicated = component.instances != 0;
    int offset = componentOffset(component);
    // we update c itself since it is not shared
    return (Context c)->
    {
      Value.Int thisNumber = (Value.Int)c.get(offset);
      int thisNumber0 = thisNumber.getValue();
      Value.Array ostate = (Value.Array)c.get(cn);
      Value.Array cstate = replicated ? (Value.Array)ostate.get(thisNumber0) : ostate;
      Value.Array cstate0 = new Value.Array(n+2*an);
      // copy state variables from local component context
      for (int i = 0; i < n; i++) 
        cstate0.set(i, c.get(1+offset+i));
      // copy buffer variables from global component context
      for (int i = 0; i < 2*an; i++)
      {
        Value cvalue = cstate.get(n+i);
        cstate0.set(n+i, cvalue);
      }
      if (replicated) 
      {
        Value.Array cstate1 = new Value.Array(ostate);
        cstate1.set(thisNumber0, cstate0);
        cstate0 = cstate1;
      }
      c.set(cn, cstate0, cname);
      return c;
    };
  }
  
  //-------------------------------------------------------------------------
  //
  // Commands
  //
  // -------------------------------------------------------------------------

  /****************************************************************************
   * Translate a send command.
   * @param command the original command.
   * @param component the receiver component.
   * @param eindex the index of the component instance.
   * @param action the action of the component.
   * @param emsg the message sent.
   * @return the semantics of the command; this semantics returns a
   * "null" context if the attempt to send the message failed due to
   * a full buffer; these contexts have to be filtered before further processing.
   ***************************************************************************/
  public ComSem SendCommand(AST.Command.SendCommand command,
    ComponentSymbol component, ExpSem eindex, ActionSymbol action, ExpSem emsg)
  {
    // the slot of the state variable for the component and
    // the indices for the action buffer and its length within that variable
    int stateSlot = Translator.toSlot(component.getStateVar());
    int n = component.getVars().length;
    int bufferIndex = n+2*action.getNumber();
    int numberIndex = bufferIndex+1;
    String name = component.ident.string;
    // we update c itself since it is not shared
    if (eindex == null)
      return Commands.apply(emsg, (SingleFunComSem.Single)(Value msg)->(Context c)->
      {
        Value.Array state = (Value.Array)c.get(stateSlot);
        Value.Array state0 = new Value.Array(state);
        Context c0 = c.set(stateSlot, state0, name);
        boolean okay = send(state0, bufferIndex, numberIndex, msg);
        if (!okay) c.setNull(true);
        return c0;
      });
    return Commands.apply(new ExpSem[] { eindex, emsg }, (Value[] vs, Context c)->
    {
      Value.Int index = (Value.Int)vs[0];
      Value msg = vs[1];
      Value.Array stateArray = (Value.Array)c.get(stateSlot);
      Value.Array stateArray0 = new Value.Array(stateArray);
      Context c0 = c.set(stateSlot, stateArray0, name);
      int index0 = index.getValue();
      if (index0 < 0 || index0 >= stateArray0.size()) 
        translator.runtimeError(command.exp, "invalid component index " + index0);
      Value.Array state = (Value.Array)stateArray0.get(index0);
      Value.Array state0 = new Value.Array(state);
      stateArray0.set(index0, state0);
      boolean okay = send(state0, bufferIndex, numberIndex, msg);
      if (!okay) c.setNull(true);
      return c0;
    });
  }

  /****************************************************************************
   * Try to send a message (may fail if message buffer is full)
   * @param state the component state.
   * @param bufferIndex the index of the send buffer in the state.
   * @param numberIndex the index of the number of elements in the send buffer.
   * @param msg the message to be sent.
   * @return true iff the message could be sent.
   ***************************************************************************/
  private static boolean send(Value.Array state, 
    int bufferIndex, int numberIndex, Value msg)
  {
    Value.Array buffer = (Value.Array)state.get(bufferIndex);
    Value.Int number = (Value.Int)state.get(numberIndex);
    int number0 = number.getValue();
    if (number0 == buffer.size()) return false;
    Value.Array buffer0 = new Value.Array(buffer);
    state.set(bufferIndex, buffer0);
    buffer0.set(number0, msg);
    state.set(numberIndex, new Value.Int(number0+1));
    return true;
  }
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------
