// ---------------------------------------------------------------------------
// Systems.java
// Generate tasks for verification of system invariants.
// $Id: Systems.java,v 1.64 2020/11/23 15:59:07 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2017-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.tasks;

import riscal.*;
import riscal.syntax.*;
import riscal.syntax.AST.*;
import riscal.syntax.AST.Declaration.*;
import riscal.syntax.AST.Expression.*;
import riscal.syntax.AST.SystemSpec.*;
import riscal.syntax.AST.Type.*;
import riscal.syntax.AST.Command.*;
import riscal.types.*;
import riscal.types.Environment.Symbol.*;
import riscal.tasks.Validation.*;
import riscal.tasks.Validation.Condition.Kind;

import java.util.*;

public class Systems
{
  /****************************************************************************
   * Generate tasks for declaration and put them into folder.
   * @param decl the declaration.
   * @param folder the folder.
   * @param counter a unique counter for the declaration.
   * @param validation validation the validation task generator.
   ***************************************************************************/
  public static void generateTasks(SharedSystem decl, TaskFolder folder,
    int counter, Validation validation)
  {
    // fills folder as side effect
    new Systems(decl, folder, counter, validation);
  }

  /****************************************************************************
   * Generate tasks for declaration and put them into folder.
   * @param decl the declaration.
   * @param folder the folder.
   * @param counter a unique counter for the declaration.
   * @param validation validation the validation task generator.
   ***************************************************************************/
  public static void generateTasks(DistributedSystem decl, TaskFolder folder,
    int counter, Validation validation)
  {
    // fills folder as side effect
    new Systems(decl,folder, counter, validation);
  }
  
  // the unique theorem counter
  private int counter;

  // the validation task generator
  private Validation validation;

  // the various task folders
  private TaskFolder specFolder;
  private TaskFolder verifyFolder;
  private TaskFolder initFolder;
  private TaskFolder actionFolder;

  /****************************************************************************
   * Generate the various subfolders within the given folder.
   * @param folder the parent folder.
   ***************************************************************************/
  private void initFolders(TaskFolder folder)
  {
    specFolder = new TaskFolder(folder, "Verify specification preconditions", 
        "Make sure all operation applications in specification are legal");
    verifyFolder = new TaskFolder(folder, "Verify specification", 
        "Verify validity of system invariants and measures");
    initFolder = new TaskFolder(folder, "Verify initialization precondtions", 
        "Make sure all operation applications in initialization are legal");
    actionFolder = new TaskFolder(folder, "Verify action preconditions", 
        "Make sure all operation applications in actions are legal");
  }

  // --------------------------------------------------------------------------
  //
  // shared systems
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Generate verification task generator for declaration and folder.
   * @param decl the declaration.
   * @param folder the folder.
   * @param counter the unique counter for the system.
   * @param validation validation the validation task generator.
   ***************************************************************************/
  private Systems(SharedSystem decl, TaskFolder folder,
    int counter, Validation validation)
  {
    this.counter = counter;
    this.validation = validation;
    initFolders(folder);
    process(decl);
  }
  
  /****************************************************************************
   * Generate verification tasks for declaration.
   * @param decl the declaration.
   ***************************************************************************/
  private void process(SharedSystem decl)
  {
    // base key for generated tasks
    SystemSymbol system = (SystemSymbol)decl.ident.getSymbol();
    String key = "_" + system.ident.string + "_" + counter;

    // process state variables and their initialization expressions
    // giving rise to preliminary preconditions for initialization actions
    List<Parameter> params = new ArrayList<Parameter>(decl.vars.length);
    List<FunctionSpec> preconditions0 = new ArrayList<FunctionSpec>();
    for (Variable var : decl.vars)
    {
      params.add(new Parameter(var.ident, var.type));
      if (var.exp == null) continue;
      Expression exp0 = expandContract(var.exp, system);
      processPreExp(exp0, initFolder,
          new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
          new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
          key + "_initPre_" + var.ident.string + "_", system, null, null, false);
      FunctionSpec spec = new FunctionSpec.RequiresSpec(
          new EqualsExp(new IdentifierExp(var.ident), exp0));
      spec.setPosition(var.getPosition());
      preconditions0.add(spec);
    }

    // process system invariants (with previous invariants as preconditions)
    List<InvariantSystemSpec> invariants = new ArrayList<InvariantSystemSpec>();
    List<FunctionSpec> preconditions = new ArrayList<FunctionSpec>();
    int sn = decl.specs.length;
    for (int i = 0; i < sn; i++)
    {
      SystemSpec spec = decl.specs[i];
      if (!(spec instanceof InvariantSystemSpec)) continue;
      InvariantSystemSpec invariant = (InvariantSystemSpec)spec;
      Expression exp0 = bindOldVars(system, expandContract(invariant.exp, system));
      InvariantSystemSpec invariant0 = new InvariantSystemSpec(exp0);
      invariant0.setPosition(invariant.getPosition());
      invariants.add(invariant0);
      processPreExp(invariant0.exp, specFolder, 
          params, preconditions, 
          new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
          key + "_invPre_" + i + "_", system, null, null, false);
      FunctionSpec precondition = new FunctionSpec.RequiresSpec(invariant0.exp);
      precondition.setPosition(invariant0.getPosition());
      preconditions.add(precondition);
    }

    // process system measures (with invariants as preconditions)
    List<DecreasesSystemSpec> measures = new ArrayList<DecreasesSystemSpec>();
    for (int i = 0; i < sn; i++)
    {
      SystemSpec spec = decl.specs[i];
      if (!(spec instanceof DecreasesSystemSpec)) continue;
      DecreasesSystemSpec spec0 = (DecreasesSystemSpec)spec;
      int en = spec0.exps.length;
      List<Expression> exps0 = new ArrayList<Expression>(en);
      for (int j = 0; j < en; j++)
      {
        Expression exp0 = bindOldVars(system, expandContract(spec0.exps[j], system));
        exps0.add(exp0);
        processPreExp(exp0, specFolder, 
            params, preconditions, 
            new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
            key + "_decPre_" + i + "_" + j + "_", system, null, null, false);
      }
      DecreasesSystemSpec spec1 = new DecreasesSystemSpec(exps0);
      spec1.setPosition(spec0.getPosition());
      measures.add(spec1);
    }

    // process initialization action (with preliminary precondition)
    Action iaction = decl.init;
    if (iaction == null)
    {
      
      Identifier init = new Identifier("_init");
      init.setSymbol(new ActionSymbol(system, 0, init, new ValueSymbol[0], 
          ActionSymbol.Kind.command));
      iaction = new Action.CommandAction(init,
          new ArrayList<Parameter>(), null, new ArrayList<Command>());
    }
    processAction(iaction, initFolder, true, params, 
        preconditions0, invariants, measures,
        key + "_initPre_");

    // process system actions (with invariants as preconditions)
    int an = decl.actions.length;
    for (int i = 0; i < an; i++)
    {
      Action action = decl.actions[i];
      processAction(action, actionFolder, false, params, 
          preconditions, invariants, measures,
          key + "_actionPre_" + i + "_");
    }
  }
  
  // --------------------------------------------------------------------------
  //
  // system actions
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Process system action, adding preconditions to task folder.
   * @param action the action.
   * @param folder the task folder.
   * @param initial true if the action is for initialization.
   * @param params the parameters for closing the tasks.
   * @param preconditions the preconditions for the parameters.
   * @param invariants the invariants of the system.
   * @param measures the measures of the system.
   * @param key the prefix to use for the generated operations.
   ***************************************************************************/
  private void processAction(Action action, TaskFolder folder, boolean initial,
    List<Parameter> params, List<FunctionSpec> preconditions,
    List<InvariantSystemSpec> invariants, List<DecreasesSystemSpec> measures,
    String key)
  {
    ActionSymbol asymbol = (ActionSymbol)action.getIdent().getSymbol();
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent(); 
    List<Parameter> params0 = actionParams(action);
    TaskFolder folder0 = folder;
    TaskFolder verifyFolder0 = verifyFolder;
    if (!initial)
    {
      String cname = csymbol == null ? "" : csymbol.ident.string + ".";
      String aname = action.getIdent().string;
      folder0 = new TaskFolder(folder0,
          "action " + cname + aname, "Are operation applications legal?");
      folder0.setSourcePosition(action.getIdent().getPosition());
      verifyFolder0 = new TaskFolder(verifyFolder0,
          "action " + cname + aname, "Does action meet specification?");
      verifyFolder0.setSourcePosition(action.getIdent().getPosition());
    }
    List<FunctionSpec> preconditions0 = csymbol == null || initial ? 
        new ArrayList<FunctionSpec>() : paramPreconditions(asymbol);
    if (action.exp != null)
    {
      Expression exp0 = expandContract(action.exp, system);
      Expression exp1 = exp0;
      processPreExp(exp1, folder0,
          params, preconditions, params0, preconditions0,
          key + "guard_", system, csymbol, null, false);
      FunctionSpec spec0 = new FunctionSpec.RequiresSpec(exp0);
      spec0.setPosition(exp0.getPosition());
      preconditions0.add(spec0);
    }
    if (action instanceof Action.CommandAction)
    {
      Action.CommandAction action0 = (Action.CommandAction)action;
      processPreAction(action0, folder0, initial,
          params, preconditions, params0, preconditions0,
          key + "cpre_");
      verifyCommand(action0, verifyFolder0, initial,
          params, preconditions, params0, preconditions0,
          invariants, measures,
          key + "cverify_");
    }
    else if (action instanceof Action.FunctionAction)
    {
      Action.FunctionAction action0 = (Action.FunctionAction)action;
      Expression exp0 = expandContract(action0.fexp, system);
      processPreExp(exp0, folder0,
          params, preconditions, params0, preconditions0,
          key + "fpre_", system, csymbol, asymbol, initial);
      processType(exp0, folder0,
          params, preconditions, params0, preconditions0,
          asymbol, key + "ftype_");
      verifyFunction(exp0, verifyFolder0, initial,
          params, preconditions, params0, preconditions0,
          invariants, measures,
          asymbol, key + "fverify_");
    }
    else if (action instanceof Action.PredicateAction)
    {
      Action.PredicateAction action0 = (Action.PredicateAction)action;
      List<Parameter> params1 = new ArrayList<Parameter>(params0);
      if (csymbol == null)
      {
        ValueSymbol[] vars0 = asymbol.system.getNewVars();
        for (ValueSymbol var0 : vars0) 
        {
          Parameter param0 = new Parameter(var0.ident, var0.type);
          params1.add(param0);
        }
      }
      else
      {
        ValueSymbol[] newVars = asymbol.system.getNewVars();
        for (ValueSymbol newVar : newVars)
        {
          Parameter param0 = new Parameter(newVar.ident, newVar.type);
          params1.add(param0);
        }
      }
      Expression exp0 = expandContract(action0.pexp, system);
      if (csymbol != null) exp0 = closeComponent(csymbol, true, exp0);
      processPreExp(exp0, folder0,
          params, preconditions, params1, preconditions0,
          key + "ppre_", system, csymbol, asymbol, initial);
      verifyPredicate(exp0, verifyFolder0, initial,
          params, preconditions, params1, preconditions0, 
          invariants, measures,
          key + "pverify_", asymbol);
    }
    else
      Main.error("unknown action " + action);
  }

  /****************************************************************************
   * Get the parameters of an action.
   * @param action the action.
   * @return the parameters.
   ***************************************************************************/
  private List<Parameter> actionParams(Action action)
  {
    ActionSymbol asymbol = (ActionSymbol)action.getIdent().getSymbol();
    ComponentSymbol csymbol = asymbol.getComponent(); 
    List<Parameter> params0 = new ArrayList<Parameter>();
    if (csymbol != null)
    {
      Identifier tident = new Identifier(TypeChecker.thisName);
      int N = csymbol.instances == 0 ? 1 : csymbol.instances;
      Type ttype = natType(N-1);
      tident.setSymbol(new ValueSymbol(tident, ttype, ValueSymbol.Kind.value));
      params0.add(new Parameter(tident, ttype));
    }
    params0.addAll(Arrays.asList(action.params));
    return params0;
  }
  
  /****************************************************************************
   * Process expression, adding type checking condition to given task folder.
   * @param exp the expression (the body of a function action)
   * @param folder the task folder for the type checking condition
   * @param params the parameters for closing the expression.
   * @param preconditions the preconditions for the parameters.
   * @param params the additional parameters for closing the expression.
   * @param preconditions the additional preconditions for the parameters.
   * @param asymbol the action in which the expression occurs.
   * @param key the prefix to use for the generated operations.
   * @param state the state parameter (for distributed systems, otherwise null)
   ***************************************************************************/
  private void processType(Expression exp, TaskFolder folder, 
    List<Parameter> params, List<FunctionSpec> preconditions, 
    List<Parameter> params0, List<FunctionSpec> preconditions0, 
    ActionSymbol asymbol, String key)
  {
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent();
    ValueSymbol[] vars = csymbol == null ?
        asymbol.system.getVars() : csymbol.getVars();
    Type type = varType(vars);
    Expression constraint = Validation.getConstraint(type, exp, exp.getType());
    if (constraint == null) return;
    TheoremParamDefinition theorem = createTheorem(
        new Multiple(false), new Identifier(key), 
        params, preconditions, params0, preconditions0, 
        closeSystem(system, constraint), system, csymbol);
    if (theorem != null)
    {
      CheckTask task = new CheckTask(folder, "Does expression type match state type?", 
          constraint.toString(),
          (FunctionSymbol)theorem.ident.getSymbol(), theorem, 
          validation.getDeclarations());
      task.setSourcePosition(exp.getPosition());
      task.setRefutation(validation.getCounterExample());
    }
  }
  
  /****************************************************************************
   * Process action command, adding preconditions to given task folder.
   * @param action the action.
   * @param folder the task folder for the preconditions.
   * @param initial true if the action is for initialization
   * @param params the parameters for closing the expression.
   * @param preconditions the preconditions for the parameters.
   * @param params0 the additional parameters for closing the expression.
   * @param preconditions0 the additional preconditions for the parameters.
   * @param key the prefix to use for the generated operations.
   ***************************************************************************/
  private void processPreAction(Action.CommandAction action, TaskFolder folder, 
    boolean initial, 
    List<Parameter> params, List<FunctionSpec> preconditions, 
    List<Parameter> params0, List<FunctionSpec> preconditions0, 
    String key)
  {
    // use mostly dummy information for precondition generator
    ActionSymbol asymbol = (ActionSymbol)action.getIdent().getSymbol();
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent();
    OpInfo info = new OpInfo(system.getFunctionSymbol().ident,
        null, null, new FunctionSpec[] { }, null, null);
    Commands commands = new Commands(info, validation.getCallers(), this);
    List<Condition> conditions = commands.def(action.commands);
    int n = conditions.size();
    for (int i = 0; i < n; i++)
    {
      Condition condition = conditions.get(i);
      String description = condition.phrase.toString();
      String name;
      switch(condition.kind)
      {
      case Constraint:
        name = "Does type constraint of operation hold?";
        break;
      case Application:
        name = "Does operation precondition hold?";
        break;
      case Select:
        name = "Is index value legal?";
        break;
      case Update:
        name = "Is update value legal?";
        break;
      case Assignment:
        name = "Is assigned value legal?";
        break;
      case Assertion:
        name = "Does assertion hold?";
        break;
      case Choice:
        name = "Is choice possible?";
        break;
      case InvariantInitial:
        name = "Does loop invariant initially hold?";
        break;
      case InvariantPreserved:
        name = "Is loop invariant preserved?";
        break;
      case LoopMeasureNonNegative:
        name = "Is loop measure non-negative?";
        break;
      case LoopMeasureDecreased:
        name = "Is loop measure decreased?";
        break;
      case FunctionMeasureDecreased:
        name = "Is function measure decreased?";
        break;
      case ResultCorrect:
        name = "Is result correct?";
        break;
      default:
        name = "Unknown task";
      }
      Identifier ident = new Identifier(key + i);
      Expression exp1 = initial || csymbol == null ? condition.exp : 
        removeMessage(asymbol, condition.exp);
      TheoremParamDefinition theorem = createTheorem(
          new Multiple(false), ident, 
          params, preconditions, params0, preconditions0, 
          closeSystem(system, exp1), system, csymbol);
      if (theorem != null)
      {
        CheckTask task = new CheckTask(folder, name, description,
            (FunctionSymbol)theorem.ident.getSymbol(), theorem, 
            validation.getDeclarations());
        task.setSourcePosition(condition.pos);
        task.addAuxiliaryPositions(condition.path);
        task.setRefutation(validation.getCounterExample());
      }
    }
  }
  
  /****************************************************************************
   * Process action commands, adding invariant and measures tasks to task folder
   * @param action the action.
   * @param folder the task folder for the invariant and measures tasks.
   * @param initial true if the action is for initialization
   * @param params the parameters for closing the expression.
   * @param preconditions the preconditions for the parameters.
   * @param params the additional parameters for closing the expression.
   * @param preconditions the additional preconditions for the parameters.
   * @param invariants the invariants.
   * @param measures the measures.
   * @param key the prefix to use for the generated operations.
   ***************************************************************************/
  private void verifyCommand(Action.CommandAction action, TaskFolder folder, boolean initial, 
    List<Parameter> params, List<FunctionSpec> preconditions, 
    List<Parameter> params0, List<FunctionSpec> preconditions0, 
    List<InvariantSystemSpec> invariants, List<DecreasesSystemSpec> measures,
    String key)
  { 
    // use mostly dummy information for precondition generator
    ActionSymbol asymbol = (ActionSymbol)action.getIdent().getSymbol();
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent();
    OpInfo info = new OpInfo(system.getFunctionSymbol().ident,
        null, null, new FunctionSpec[] { }, null, null);
    Commands commands = new Commands(info, validation.getCallers(), this);

    // the generated conditions
    List<Condition> conditions = new ArrayList<Condition>();
   
    // generate invariant conditions
    if (csymbol == null || !initial)
    {
      for (InvariantSystemSpec invariant : invariants)
      {
        Expression exp = closeSystem(system, invariant.exp);
        if (csymbol != null) exp = updateComponent(csymbol, exp);
        Kind kind = initial ? Kind.InvariantInitial : Kind.InvariantPreserved;
        Condition invariant0 = new Condition(kind, invariant, exp);
        conditions.addAll(commands.process(action.commands, invariant0));
      }
    }
    
    // the measures and their weakest preconditions
    for (DecreasesSystemSpec measure : measures)
    {
      if (initial)
      {
        Condition measure0 = notNegativeCondition(measure);
        measure0 = measure0.clone(closeSystem(system, measure0.exp));
        if (csymbol != null) 
          measure0 = measure0.clone(updateComponent(csymbol, measure0.exp));
        conditions.addAll(commands.process(action.commands, measure0));
      }
      else
      {
        Condition measure0 = decreasesCondition(measure);
        measure0 = measure0.clone(closeSystem(system, measure0.exp));
        if (csymbol != null) 
          measure0 = measure0.clone(updateComponent(csymbol, measure0.exp));
        List<Condition> measures1 = commands.process(action.commands, measure0);
        List<Binder> binders = getMeasureBinders(measure, true);
        for (Condition measure1 : measures1)
        {
          Expression mexp0 = new LetParExp(binders, measure1.exp);
          mexp0 = closeSystem(system, mexp0);
          conditions.add(measure1.clone(mexp0));
        }
      }
    }
    
    // translate conditions to tasks
    generateActionTasks(asymbol, initial, conditions, 
        folder, params, preconditions, params0, preconditions0, key);
  } 
  
  /***************************************************************************
   * Create non-negativity condition for measure.
   * @param measure the measure.
   * @return the condition.
   *************************************************************************/
  private static Condition notNegativeCondition(DecreasesSystemSpec measure)
  {
    List<Binder> binders = getMeasureBinders(measure, true);
    Expression exp = null;
    for (Binder binder : binders)
    {
      exp = AST.and(exp, 
          new GreaterEqualExp(new IdentifierExp(binder.ident), 
              new NumberLiteralExp(AST.decimal(0))));
    }
    exp = new LetParExp(binders, exp);
    return new Condition(Kind.LoopMeasureNonNegative, measure, exp);
  }
  
  /***************************************************************************
   * Create decrement condition for measure.
   * @param measure the measure.
   * @return the condition.
   *************************************************************************/
  private static Condition decreasesCondition(DecreasesSystemSpec measure)
  {
    List<Binder> binders = getMeasureBinders(measure, true);
    List<Binder> nbinders = getMeasureBinders(measure, false);
    Expression exp = null;
    Expression equalities = null;
    int n = binders.size();
    for (int i = 0; i < n; i++)
    {
      Identifier oldid = binders.get(i).ident;
      Identifier newid = nbinders.get(i).ident;
      Expression inequality = 
          new LessExp(new IdentifierExp(newid), new IdentifierExp(oldid));
      Expression conjunction = AST.and(equalities, inequality);
      exp = AST.or(exp, conjunction);
      Expression equality =
          new EqualsExp(new IdentifierExp(newid), new IdentifierExp(oldid));
      equalities = AST.and(equalities, equality);
    }
    exp = new LetParExp(nbinders, exp);
    return new Condition(Kind.LoopMeasureDecreased, measure, exp);
  }
  
  /****************************************************************************
   * Generate theorem checking tasks from conditions.
   * @param asymbol the action from which the task emerges.
   * @param initial true iff the action is initial.
   * @param conditions the conditions.
   * @param folder the folder where to create the tasks.
   * @param params the parameters of the conditions.
   * @param preconditions the preconditions of these parameters.
   * @param params0 additional parameters of the conditions.
   * @param preconditions0 the preconditions of these parameters.
   * @param key the key to be used as a basis of the theorem names.
   ***************************************************************************/
  private void generateActionTasks(ActionSymbol asymbol, boolean initial,
    List<Condition> conditions, TaskFolder folder, List<Parameter> params, 
    List<FunctionSpec> preconditions, List<Parameter> params0, 
    List<FunctionSpec> preconditions0, String key)
  {
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent();
    int n = conditions.size();
    for (int i = 0; i < n; i++)
    {
      Condition condition = conditions.get(i);
      Identifier ident = new Identifier(key+i);
      Expression exp = initial || csymbol == null ? condition.exp : 
        removeMessage(asymbol, condition.exp);
      TheoremParamDefinition theorem = createTheorem(
          new Multiple(false), ident, 
          params, preconditions, params0, preconditions0,
          exp, system, csymbol);
      if (theorem != null)
      {
        CheckTask task = new CheckTask(folder, taskName(condition.kind), 
            condition.phrase.toString(),
            (FunctionSymbol)theorem.ident.getSymbol(), theorem, 
            validation.getDeclarations());
        task.setSourcePosition(condition.pos);
        task.addAuxiliaryPositions(condition.path);
        task.setRefutation(validation.getCounterExample());
      }
    }
  }
  
  /****************************************************************************
   * Process expression, adding preconditions to given task folder.
   * @param exp the expression.
   * @param folder the task folder for the preconditions.
   * @param params the parameters for closing the expression.
   * @param preconditions the preconditions for the parameters.
   * @param params the additional parameters for closing the expression.
   * @param preconditions the additional preconditions for the parameters.
   * @param key the prefix to use for the generated operations.
   * @param system the current system.
   * @param csymbol the current component (if any, otherwise null)
   * @param asymbol the current action symbol (if any, otherwise null)
   * @param initial true iff the action is initial
   ***************************************************************************/
  private void processPreExp(Expression exp, TaskFolder folder, 
    List<Parameter> params, List<FunctionSpec> preconditions, 
    List<Parameter> params0, List<FunctionSpec> preconditions0,
    String key, SystemSymbol system, ComponentSymbol csymbol, ActionSymbol asymbol,
    boolean initial)
  {
    // use dummy (system execution) function as argument for precondition generator
    List<Condition> conditions = Preconditions.generate(exp,
        system.getFunctionSymbol(), validation.getCallers());   
    int n = conditions.size();
    for (int i = 0; i < n; i++)
    {
      Condition condition = conditions.get(i);
      String name;
      String description;
      switch(condition.kind)
      {
      case Assertion:
        name = "Does assertion hold?";
        description = condition.phrase.toString();
        break;
      case Choice:
        name = "Is choice possible?";
        description = condition.phrase.toString();
        break;
      case Select:
        name = "Is index value legal?";
        description = condition.phrase.toString();
        break;
      case Update:
        name = "Is update value legal?";
        description = condition.phrase.toString();
        break;
      default:
        name = "Does operation precondition hold?";
        description = condition.phrase.toString();
      }
      Identifier ident = new Identifier(key + i);
      Expression exp0 = initial || asymbol == null ? condition.exp : 
        removeMessage(asymbol, condition.exp);
      TheoremParamDefinition theorem = createTheorem(
          new Multiple(false), ident, 
          params, preconditions, params0, preconditions0,
          closeSystem(system, exp0), system, csymbol);
      if (theorem != null)
      {
        CheckTask task = new CheckTask(folder, name, description,
            (FunctionSymbol)theorem.ident.getSymbol(), theorem, 
            validation.getDeclarations());
        task.setSourcePosition(condition.pos);
        task.addAuxiliaryPositions(condition.path);
        task.setRefutation(validation.getCounterExample());
      }
    }
  }


  /****************************************************************************
   * Process function expression, adding invariant and measure tasks to folder.
   * @param exp the expression (the body of a function action).
   * @param folder the task folder for the tasks.
   * @param initial true if the action is for initialization.
   * @param params the parameters for closing the expression.
   * @param preconditions the preconditions for the parameters.
   * @param invariants the invariants.
   * @param preconditions0 the additional preconditions for the parameters.
   * @param invariants0 the additional invariants.
   * @param measures the measures.
   * @param asymbol the action in which the expression occurs.
   * @param key the prefix to use for the generated operations.
   * @param state the state parameter (for distributed systems, otherwise null)
   ***************************************************************************/
  private void verifyFunction(Expression exp, TaskFolder folder, boolean initial, 
    List<Parameter> params, List<FunctionSpec> preconditions, 
    List<Parameter> params0, List<FunctionSpec> preconditions0, 
    List<InvariantSystemSpec> invariants, 
    List<DecreasesSystemSpec> measures, ActionSymbol asymbol, 
    String key)
  {
    List<Condition> conditions = new ArrayList<Condition>();
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent();
    ValueSymbol[] vars = system.getVars();
    
    // generate invariant conditions
    if (csymbol == null && !initial)
    {
      for (InvariantSystemSpec invariant : invariants)
      {
        Expression body = csymbol == null ? bindTuple(vars, exp, invariant.exp) :
          bindComponent(csymbol, exp, invariant.exp);   
        body = closeSystem(system, body);
        Kind kind = initial ? Kind.InvariantInitial : Kind.InvariantPreserved;
        Condition condition = new Condition(kind, invariant, body);
        condition.addPosition(exp.getPosition());
        conditions.add(condition);
      }
    }
    
    // generate measure conditions
    for (DecreasesSystemSpec measure : measures)
    {
      if (initial)
      {
        Condition measure0 = notNegativeCondition(measure);
        measure0 = measure0.clone(closeSystem(system, measure0.exp));
        measure0.addPosition(exp.getPosition());
        Expression mexp = csymbol == null ? bindTuple(vars, exp, measure0.exp) :
          bindComponent(csymbol, exp, measure0.exp);
        conditions.add(measure0.clone(mexp));
      }
      else
      {
        Condition measure0 = decreasesCondition(measure);
        measure0 = measure0.clone(closeSystem(system, measure0.exp));
        measure0.addPosition(exp.getPosition());
        Expression mexp = csymbol == null ? bindTuple(vars, exp, measure0.exp) :
          bindComponent(csymbol, exp, measure0.exp);
        List<Binder> binders = getMeasureBinders(measure, true);
        conditions.add(measure0.clone(closeSystem(system, new LetParExp(binders, mexp))));
      }
    }
    
    // turn conditions into tasks
    generateActionTasks(asymbol, initial, conditions, 
        folder, params, preconditions, params0, preconditions0, key);
  }
  
 
  
  /****************************************************************************
   * Process predicate expression, adding invariant and measure tasks to folder.
   * @param exp the expression (the body of a predicate action).
   * @param folder the task folder for the tasks.
   * @param initial true if the action is for initialization.
   * @param params the parameters for closing the expression.
   * @param preconditions the preconditions for the parameters.
   * @param params the additional parameters for closing the expression.
   * @param preconditions the additional preconditions for the parameters.
   * @param invariants the invariants.
   * @param measures the measures.
   * @param key the prefix to use for the generated operations.
   * @param asymbol the action.
   ***************************************************************************/
  private void verifyPredicate(Expression exp, TaskFolder folder, boolean initial, 
    List<Parameter> params, List<FunctionSpec> preconditions, 
    List<Parameter> params0, List<FunctionSpec> preconditions0, 
    List<InvariantSystemSpec> invariants, List<DecreasesSystemSpec> measures, 
    String key, ActionSymbol asymbol)
  {
    SystemSymbol system = asymbol.system;
    ComponentSymbol csymbol = asymbol.getComponent();
    List<Condition> conditions = new ArrayList<Condition>();
    ValueSymbol[] vars = asymbol.system.getVars();
    ValueSymbol[] newVars = asymbol.system.getNewVars();
    
    // generate invariant conditions
    if (csymbol == null || !initial)
    {
      for (InvariantSystemSpec invariant : invariants)
      {
        Expression iexp = closeSystem(system, invariant.exp);
        iexp = new ImpliesExp(exp, bind(vars, newVars, iexp));
        iexp = closeSystem(system, iexp);
        Kind kind = initial ? Kind.InvariantInitial : Kind.InvariantPreserved;
        Condition condition = new Condition(kind, invariant, iexp);
        condition.addPosition(exp.getPosition());
        conditions.add(condition);
      }
    }
    
    // generate measure conditions
    for (DecreasesSystemSpec measure : measures)
    {
      if (initial)
      {
        Condition measure0 = notNegativeCondition(measure);
        measure0.addPosition(exp.getPosition());
        Expression mexp = new ImpliesExp(exp, bind(vars, newVars, measure0.exp));
        mexp = closeSystem(system, mexp);
        conditions.add(measure0.clone(mexp));
      }
      else
      {
        Condition measure0 = decreasesCondition(measure);
        measure0 = measure0.clone(closeSystem(system, measure0.exp));
        measure0.addPosition(exp.getPosition());
        List<Binder> binders = getMeasureBinders(measure, true);
        Expression mexp = new ImpliesExp(exp, 
            new LetParExp(binders, bind(vars, newVars, measure0.exp)));
        mexp = closeSystem(system, mexp);
        conditions.add(measure0.clone(mexp));
      }
    }
    
    // turn conditions into tasks
    generateActionTasks(asymbol, initial, conditions, 
        folder, params, preconditions, params0, preconditions0, key);
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliaries
  //
  // --------------------------------------------------------------------------
   
  /***************************************************************************
   * Create, optimize, and process theorem
   * @param multiple does theorem have non-deterministic semantics?
   * @param ident its identifier
   * @param params its parameters
   * @param preconditions their preconditions
   * @param params0 additional parameters
   * @param preconditions0 their preconditions
   * @param body its body
   * @param system the system
   * @param component the current component (if any, else null)
   * @return the (optimized, processed and translated) theorem 
   *         (null, if something went wrong)
   *         sets as a side effect "counterExample" to a predicate 
   *         that attempts to refute the theorem
   **************************************************************************/
  private TheoremParamDefinition createTheorem(Multiple multiple, Identifier ident, 
    List<Parameter> params, List<FunctionSpec> preconditions,
    List<Parameter> params0, List<FunctionSpec> preconditions0,
    Expression formula, SystemSymbol system, ComponentSymbol csymbol)
  {
    Expression exp = null;
    for (FunctionSpec pre0 : preconditions0)
    {
      FunctionSpec.RequiresSpec pre1 = (FunctionSpec.RequiresSpec)pre0;
      exp = AST.and(exp, pre1.exp);
    }
    if (exp != null) 
    {
      formula = new ImpliesExp(exp, formula);
      formula.setType(new BoolType());
    }
    if (csymbol != null) formula = closeComponent(csymbol, false, formula);
    if (!params0.isEmpty())
    {
      List<QuantifiedVariableCore> qvcore = new ArrayList<QuantifiedVariableCore>();
      for (Parameter param0 : params0)
      {
        ValueSymbol symbol = (ValueSymbol)param0.ident.getSymbol();
        QuantifiedVariableCore qvc = 
            new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(symbol.ident, symbol.type);
        qvcore.add(qvc);
      }
      formula = new ForallExp(new QuantifiedVariable(qvcore, null), formula);
      formula.setType(new BoolType());
    }
    if (system.getComponents() != null)
    {
      List<FunctionSpec> preconditions1 = new ArrayList<FunctionSpec>();
      for (FunctionSpec pre : preconditions)
      {
        FunctionSpec.RequiresSpec pre0 = (FunctionSpec.RequiresSpec)pre;
        preconditions1.add(new FunctionSpec.RequiresSpec(closeSystem(system, pre0.exp)));
      }
      preconditions = preconditions1;
    }
    return validation.createTheorem(multiple, ident, params, preconditions, formula);
  }
  
  /****************************************************************************
   * Expand contracts for function applications in expression.
   * @param exp the original expression.
   * @param system the system in which the expression appears.
   * @return the expanded expression (with position preserved).
   ***************************************************************************/
  private Expression expandContract(Expression exp, SystemSymbol system)
  {
    // use dummy (system execution) function for contract expansion
    Expression exp0 = UseContracts.execute(exp, 
        system.getFunctionSymbol(), validation.getCallers());
    exp0.setPosition(exp.getPosition());
    return exp0;
  }
  
  /****************************************************************************
   * Give name of task for condition of a certain kind.
   * @param kind the condition.
   * @return the task name.
   ***************************************************************************/
  private static String taskName(Kind kind)
  {
    switch (kind)
    {
    case InvariantInitial: return "Does system invariant initially hold?";
    case InvariantPreserved: return "Is system invariant preserved?";
    case LoopMeasureNonNegative: return "Is system measure initially non-negative?";
    case LoopMeasureDecreased: return "Is system measure decreased?";
    default: return "unknown task";
    }
  }
  
  /****************************************************************************
   * Get name of measure
   * @param i the number of the measure
   * @param old if true, name of old measure, otherwise name of new one
   * @return its name
   ***************************************************************************/
  private static String getMeasureName(int i, boolean old)
  {
    String key = old ? "_measure" : "_nmeasure" ;
    if (i == 0) return key;
    return key + i;
  }
  
  /****************************************************************************
   * Get list of binders for measure.
   * @param measure the measure.
   * @param old iff true, binders for old values, otherwise for new ones.
   * @return the binders.
   ***************************************************************************/
  private static List<Binder> getMeasureBinders(DecreasesSystemSpec measure,
    boolean old)
  {
    Expression[] measures = measure.exps;
    int n = measures.length;
    List<Binder> binders = new ArrayList<Binder>();
    for (int i=0; i<n; i++)
    {
      String name = getMeasureName(i, old);
      binders.add(new Binder(new Identifier(name), measures[i]));
    }
    return binders;     
  }
  
  /****************************************************************************
   * Get type of system variables.
   * @param vars the system variables
   * @return a tuple type with the variable types as components (but in case 
   * of a a single variable, the type of the variable itself).
   ***************************************************************************/
  private static Type varType(ValueSymbol[] vars)
  {
    if (vars.length == 1) return vars[0].type;
    List<Type> types = new ArrayList<Type>();
    for (ValueSymbol var : vars)
      types.add(var.type);
    return new TupleType(types);
  }
  
  // --------------------------------------------------------------------------
  //
  // distributed systems
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Generate verification task generator for declaration and folder.
   * @param decl the declaration.
   * @param folder the folder.
   * @param counter the unique counter for the system.
   * @param validation validation the validation task generator.
   ***************************************************************************/
  private Systems(DistributedSystem decl, TaskFolder folder,
    int counter, Validation validation)
  {
    this.counter = counter;
    this.validation = validation;
    initFolders(folder);
    process(decl);
  }
   
  /****************************************************************************
   * Generate verification tasks for declaration.
   * @param decl the declaration.
   ***************************************************************************/
  private void process(DistributedSystem decl)
  {
    // base key for generated tasks
    SystemSymbol system = (SystemSymbol)decl.ident.getSymbol();
    String key = "_" + system.ident.string + "_" + counter;

    // the component states as parameters
    List<ValueSymbol> vars = new ArrayList<ValueSymbol>();
    List<ValueSymbol> newVars = new ArrayList<ValueSymbol>();
    List<Parameter> params = new ArrayList<Parameter>();
    for (ComponentSymbol csymbol : system.getComponents())
    {
      ValueSymbol vsymbol = componentVariable(csymbol, false, false);
      ValueSymbol vsymbol0 = componentVariable(csymbol, false, true);
      vars.add(vsymbol);
      newVars.add(vsymbol0);
      params.add(new Parameter(vsymbol.ident, vsymbol.type));
    }
    system.setVars(vars.toArray(new ValueSymbol[vars.size()]));
    system.setNewVars(newVars.toArray(new ValueSymbol[newVars.size()]));
    
    // process initialization expressions
    // giving rise to preliminary preconditions for initialization actions
    List<FunctionSpec> preconditions0 = bufferPreconditions(system);
    for (Component component : decl.components)
    {
      ComponentSymbol csymbol = (ComponentSymbol)component.ident.getSymbol();
      for (Variable var : component.vars)
      {
        if (var.exp == null) continue;
        Expression exp0 = expandContract(var.exp, system);
        processPreExp(exp0, initFolder,
            new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
            new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
            key + "_initPre_" + var.ident.string + "_", system, null, null, false);
        Identifier init = new Identifier("_init");
        List<Binder> binders = new ArrayList<Binder>();
        binders.add(new Binder(init, exp0));
        exp0 = new EqualsExp(new IdentifierExp(var.ident), new IdentifierExp(init));
        exp0 = closeComponent(csymbol, false, exp0);
        exp0 = new LetExp(binders, exp0);
        if (csymbol.instances != 0)
        {
          ValueSymbol thisSymbol = componentIndex(csymbol,TypeChecker.thisName);
          List<QuantifiedVariableCore> qvc = new ArrayList<QuantifiedVariableCore>();
          qvc.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(thisSymbol.ident, thisSymbol.type));
          exp0 = new ForallExp(new QuantifiedVariable(qvc, null), exp0);
        }
        FunctionSpec spec = new FunctionSpec.RequiresSpec(exp0);
        spec.setPosition(var.getPosition());
        preconditions0.add(spec);
      }
    }
    
    // process system invariants (with previous invariants as preconditions)
    List<InvariantSystemSpec> invariants = new ArrayList<InvariantSystemSpec>();
    List<Condition> invariants0 = new ArrayList<Condition>();
    List<FunctionSpec> preconditions = new ArrayList<FunctionSpec>();
    int sn = decl.specs.length;
    for (int i = 0; i < sn; i++)
    {
      SystemSpec spec = decl.specs[i];
      if (!(spec instanceof InvariantSystemSpec)) continue;
      InvariantSystemSpec invariant = (InvariantSystemSpec)spec;
      Expression exp0 = bindOldVars(system,expandContract(invariant.exp, system));
      InvariantSystemSpec invariant0 = new InvariantSystemSpec(exp0);
      invariant0.setPosition(invariant.getPosition());
      invariants.add(invariant0);
      invariants0.add(new Condition(Kind.InvariantInitial, invariant, 
          closeSystem(system, invariant0.exp)));
      processPreExp(invariant0.exp, specFolder, 
          params, preconditions, 
          new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
          key + "_invPre_" + i + "_", system, null, null, false);
      FunctionSpec precondition = new FunctionSpec.RequiresSpec(invariant0.exp);
      precondition.setPosition(invariant0.getPosition());
      preconditions.add(precondition);
    }
    
    // process system measures (with invariants as preconditions)
    List<DecreasesSystemSpec> measures = new ArrayList<DecreasesSystemSpec>();
    for (int i = 0; i < sn; i++)
    {
      SystemSpec spec = decl.specs[i];
      if (!(spec instanceof DecreasesSystemSpec)) continue;
      DecreasesSystemSpec spec0 = (DecreasesSystemSpec)spec;
      int en = spec0.exps.length;
      List<Expression> exps0 = new ArrayList<Expression>(en);
      for (int j = 0; j < en; j++)
      {
        Expression exp0 = bindOldVars(system, expandContract(spec0.exps[j], system));
        exps0.add(exp0);
        processPreExp(exp0, specFolder, 
            params, preconditions, 
            new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
            key + "_decPre_" + i + "_" + j + "_", system, null, null, false);
      }
      DecreasesSystemSpec spec1 = new DecreasesSystemSpec(exps0);
      spec1.setPosition(spec0.getPosition());
      measures.add(spec1);
    }
    
    // process invariants with respect to initialization actions of components
    int cn = decl.components.length;
    for (int i = cn-1; i >= 0; i--)
    {
      Component component = decl.components[i];
      if (component.init == null) continue;
      ComponentSymbol csymbol = (ComponentSymbol)component.ident.getSymbol();
      int in = csymbol.instances;
      if (in == 0)
      {
        List<Condition> invariants1 = preAction(component.init, invariants0);
        invariants0 = closeComponent(csymbol, invariants1);
        continue;
      }
      ValueSymbol thisSymbol = componentIndex(csymbol,TypeChecker.thisName);
      List<Condition> invariants1 = invariants0;
      for (int j = in-1; j >= 0; j--)
      {
        invariants1 = preAction(component.init, invariants1);
        invariants1 = closeComponent(csymbol, invariants1);
        List<Condition> invariants2 = new ArrayList<Condition>();
        for (Condition invariant1 : invariants1)
        {
          Expression exp1 = invariant1.exp;
          List<Binder> binders = new ArrayList<Binder>();
          Expression iexp = new NumberLiteralExp(new Decimal(Integer.toString(j)));
          iexp.setType(new IntType(j,j));
          binders.add(new Binder(thisSymbol.ident, iexp));
          Expression exp2 = new LetExp(binders, exp1);
          exp2.setPosition(exp1.getPosition());
          exp2.setType(exp1.getType());
          invariants2.add(invariant1.clone(exp2));
        }
        invariants1 = invariants2;
      }
      invariants0 = invariants1;
    }
    int in = invariants0.size();
    for (int i = 0; i < in; i++)
    {
      Condition invariant = invariants0.get(i);
      Identifier ident = new Identifier(key + "_initInv_" + i);
      TheoremParamDefinition theorem = createTheorem(
          new Multiple(false), ident, 
          params, preconditions0, new ArrayList<Parameter>(), new ArrayList<FunctionSpec>(),
          closeSystem(system, invariant.exp), system, null);
      if (theorem != null)
      {
        CheckTask task = new CheckTask(verifyFolder, taskName(invariant.kind), 
            invariant.phrase.toString(),
            (FunctionSymbol)theorem.ident.getSymbol(), theorem, 
            validation.getDeclarations());
        task.setSourcePosition(invariant.pos);
        task.addAuxiliaryPositions(invariant.path);
        task.setRefutation(validation.getCounterExample());
      }
    }
    
    // process system actions of components
    for (Component component : decl.components)
    {
      String key0 = key + "_" + component.ident.string;
      
      // process initialization action (with preliminary precondition)
      if (component.init != null)
      {
        processAction(component.init, initFolder, true, params, 
            preconditions0, invariants, measures,
            key0 + "_initPre_");
      }

      // process system actions (with invariants as preconditions)
      int an = component.actions.length;
      for (int i = 0; i < an; i++)
      {
        Action action = component.actions[i];
        processAction(action, actionFolder, false, params, 
            preconditions, invariants, measures,
            key0 + "_actionPre_" + i + "_");
      }
    }
  }

  /****************************************************************************
   * Get preconditions of action with respect to postconditions.
   * @param action the action.
   * @param conditions the postconditions.
   * @return the preconditions.
   ***************************************************************************/
  private List<Condition> preAction(Action action, List<Condition> conditions)
  {
    ActionSymbol asymbol = (ActionSymbol)action.getIdent().getSymbol();
    if (action instanceof Action.CommandAction)
    {
      Action.CommandAction action0 = (Action.CommandAction)action;
      return preCommand(action0, conditions);
    }
    else if (action instanceof Action.FunctionAction)
    {
      Action.FunctionAction action0 = (Action.FunctionAction)action;
      return preFunction(action0.exp, asymbol, conditions);
    } 
    else if (action instanceof Action.PredicateAction)
    {
      Action.PredicateAction action0 = (Action.PredicateAction)action;
      return prePredicate(action0.exp, asymbol, conditions);
    }
    Main.error("Unknown action: " + action);
    return null;
  }
    
  /****************************************************************************
   * Get preconditions of action for given postconditions.
   * @param action the action.
   * @param conditions the conditions.
   * @return the preconditions.
   ***************************************************************************/
  private List<Condition> preCommand(Action.CommandAction action,
    List<Condition> conditions)
  {
    ActionSymbol asymbol = (ActionSymbol)action.getIdent().getSymbol();
    ComponentSymbol csymbol = asymbol.getComponent();
    OpInfo info = new OpInfo(asymbol.system.getFunctionSymbol().ident,
        null, null, new FunctionSpec[] { }, null, null);
    Commands commands = new Commands(info, validation.getCallers(), this);
    List<Condition> result = new ArrayList<Condition>();
    for (Condition condition : conditions)
    {
      Expression exp = condition.exp;
      if (csymbol != null) exp = updateComponent(csymbol, exp);
      Condition condition0 = condition.clone(exp);
      result.addAll(commands.process(action.commands, condition0));
    }
    return result;
  }
  
  /****************************************************************************
   * Get precondition of function expression with respect to postconditions.
   * @param exp the function expression.
   * @param asymbol the action in which the expression appears.
   * @param conditions the postconditions.
   * @return the preconditions.
   ***************************************************************************/
  private List<Condition> preFunction(Expression exp, ActionSymbol asymbol,
    List<Condition> conditions)
  {
    ComponentSymbol csymbol = asymbol.getComponent();
    SystemSymbol system = asymbol.system;
    ValueSymbol[] vars = system.getVars();
    List<Condition> result = new ArrayList<Condition>();
    for (Condition condition : conditions)
    {
      Expression body = csymbol == null ? bindTuple(vars, exp, condition.exp) :
        bindComponent(csymbol, exp, condition.exp);        
      Condition condition0 = condition.clone(body);
      condition0.addPosition(exp.getPosition());
      result.add(condition0);
    }
    return result;
  }
  
  /****************************************************************************
   * Get precondition of predicate expression with respect to postconditions.
   * @param exp the predicate expression.
   * @param asymbol the action in which the expression appears.
   * @param conditions the postconditions.
   * @return the preconditions.
   ***************************************************************************/
  private List<Condition> prePredicate(Expression exp, ActionSymbol asymbol,
    List<Condition> conditions)
  {
    List<Condition> result = new ArrayList<Condition>();
    ValueSymbol[] vars = asymbol.system.getVars();
    ValueSymbol[] newVars = asymbol.system.getNewVars();
    for (Condition condition : conditions)
    {
      Expression iexp = new ImpliesExp(exp, bind(vars, newVars, condition.exp));
      Condition condition0 = condition.clone(iexp);
      condition0.addPosition(exp.getPosition());
      result.add(condition0);
    }
    return result;
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliaries
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Get integer type with expression bound.
   * @param to the upper bound.
   * @return the type.
   ***************************************************************************/
  private static Type natType(int to)
  {
    // return new NatType(new NumberLiteralExp(new Decimal(Integer.toString(to))));
    Type type = new IntType(0, to);
    type.setCanonical();
    return type;
  }
  
  /****************************************************************************
   * Get the type of the state of a (possibly replicated) component.
   * @param csymbol the component symbol.
   * @return the component type (an array type, if the component is replicated).
   ***************************************************************************/
  private static Type componentType(ComponentSymbol csymbol)
  {
    // state variables
    List<ValueSymbol> vars = new ArrayList<ValueSymbol>(Arrays.asList(csymbol.getVars()));
    // actions
    for (ActionSymbol asymbol : csymbol.getActions())
    {
      Identifier ident = asymbol.ident;
      // message buffer content
      int B = asymbol.bsize;
      Type rtype = asymbol.params.length == 0 ? new UnitType() : 
        new RecordType(Arrays.asList(asymbol.params), true);
      rtype.setCanonical();
      Type mtype = new MapType(natType(B-1), rtype);
      mtype.setCanonical();
      Identifier msgid = new Identifier(ident.string);
      ValueSymbol msymbol = new ValueSymbol(msgid, mtype, ValueSymbol.Kind.value);
      msgid.setSymbol(msymbol);
      vars.add(msymbol);
      // message buffer size
      Identifier numid = TypeChecker.numberIdentifier(ident);
      ValueSymbol vsymbol = new ValueSymbol(numid, natType(B), ValueSymbol.Kind.value);
      numid.setSymbol(vsymbol);
      vars.add(vsymbol);
    }
    Type type = vars.isEmpty() ? new UnitType() : new RecordType(vars, true);
    type.setCanonical();
    // replicated or not?
    int N = csymbol.instances;
    if (N == 0) return type;
    Type result = new MapType(natType(N-1), type);
    result.setCanonical();
    return result;
  }
  
  /****************************************************************************
   * Get variable for component state.
   * @param csymbol the component.
   * @param original true iff the variable name without underscore shall be used.
   * @param newv true iff the variable shall refer to the post-state.
   * @return the symbol for the variable
   ***************************************************************************/
  private static ValueSymbol componentVariable(ComponentSymbol csymbol, 
    boolean original, boolean newv)
  {
    String string = csymbol.ident.string;
    if (!original) string = "_" + string;
    Identifier cident = new Identifier(string);
    if (newv) cident = TypeChecker.postIdentifier(cident);
    Type ctype = componentType(csymbol);
    ValueSymbol vsymbol = new ValueSymbol(cident, ctype, ValueSymbol.Kind.value);
    cident.setSymbol(vsymbol);
    return vsymbol;
  }
  
  /****************************************************************************
   * Generate preconditions that all action buffers are empty.
   * @param csymbols the distributed system containing the actions.
   * @return the precondition list.
   ***************************************************************************/
  private List<FunctionSpec> bufferPreconditions(SystemSymbol system)
  {
    ComponentSymbol[] csymbols = system.getComponents();
    List<FunctionSpec> result = new ArrayList<FunctionSpec>();
    Expression condition = null;
    for (ComponentSymbol csymbol : csymbols)
    {
      ValueSymbol index = csymbol.instances == 0 ? null : 
        componentIndex(csymbol, TypeChecker.thisName); 
      Expression cexp = componentExp(csymbol, index, false, false);
      Expression condition0 = null;
      for (ActionSymbol asymbol : csymbol.getActions())
      {
        ValueSymbol number = TypeChecker.actionNumber(asymbol);
        Expression nexp = new RecordSelectionExp(cexp, number.ident);
        nexp.setType(number.type);
        Expression zero = new NumberLiteralExp(new Decimal("0"));
        Type ztype = new IntType(0,0);
        ztype.setCanonical();
        zero.setType(ztype);
        Expression equalsExp = new EqualsExp(nexp, zero);
        equalsExp.setType(new BoolType());
        condition0 = AST.and(condition0, equalsExp);
        condition0.setType(new BoolType());
      }
      if (index != null)
      {
        List<QuantifiedVariableCore> qvc = new ArrayList<QuantifiedVariableCore>();
        qvc.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(index.ident, index.type));
        QuantifiedVariable qvar = new QuantifiedVariable(qvc, null);
        condition0 = new ForallExp(qvar, condition0);
        condition0.setType(new BoolType());
      }
      condition = AST.and(condition, condition0);
      condition.setType(new BoolType());
    }
    if (condition != null) result.add(new FunctionSpec.RequiresSpec(condition));
    return result;
  }
  
  /****************************************************************************
   * Get the preconditions for an action in a component.
   * @param asymbol an action appearing in a component.
   * @return the list of preconditions arising from the fact that the
   * parameter values are derived from the first message in the action buffer.
   ****************************************************************************/
  private List<FunctionSpec> paramPreconditions(ActionSymbol asymbol)
  {
    // C[this]
    ComponentSymbol csymbol = asymbol.getComponent();
    ValueSymbol componentIndex = csymbol.instances == 0 ? null : 
      componentIndex(csymbol, TypeChecker.thisName); 
    Expression cexp = componentExp(csymbol, componentIndex, false, false);
    
    // C[this].action_number > 0
    ValueSymbol number = TypeChecker.actionNumber(asymbol);
    Expression nexp = new RecordSelectionExp(cexp, number.ident);
    nexp.setType(number.type);
    Expression zero = new NumberLiteralExp(new Decimal("0"));
    Type ztype = new IntType(0,0);
    ztype.setCanonical();
    zero.setType(ztype);
    Expression greaterExp = new GreaterExp(nexp, zero);
    greaterExp.setType(new BoolType());
    
    // C[this].action[0]
    ValueSymbol buffer = TypeChecker.actionBuffer(asymbol);
    Expression bexp = new RecordSelectionExp(cexp, buffer.ident);
    MapType btype = (MapType)buffer.type;
    bexp.setType(btype);
    Expression mexp = new MapSelectionExp(bexp, zero);
    mexp.setType(btype.type2);
    
    // record <p1:p1,...,pn:pn>
    List<ExpressionIdentifier> eids = new ArrayList<ExpressionIdentifier>();
    for (ValueSymbol param : asymbol.params)
    {
      Expression pexp = new IdentifierExp(param.ident);
      pexp.setType(param.type);
      eids.add(new ExpressionIdentifier(param.ident, pexp));
    }
    Expression message = eids.isEmpty() ? new UnitExp() : new RecordExp(eids);
    message.setType(btype.type2);
    
    // C[this].action[0] = <p1:p1,...,pn:pn>
    Expression equalsExp = new EqualsExp(mexp, message);
    equalsExp.setType(new BoolType());
    
    // the resulting conjunction
    Expression conjunction = new AndExp(greaterExp, equalsExp);
    conjunction.setType(new BoolType());
    List<FunctionSpec> result = new ArrayList<FunctionSpec>();
    result.add(new FunctionSpec.RequiresSpec(conjunction));
    return result;
  }
  
  // --------------------------------------------------------------------------
  //
  // the various binding operations
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Create bound expression.
   * @param vars the variables to be bound.
   * @param vals the values of the variables.
   * @param body the body of the expression.
   * @return the bound expression.
   ***************************************************************************/
  private static Expression bind(ValueSymbol[] vars, ValueSymbol[] vals, Expression body)
  {
    List<Binder> binders = new ArrayList<Binder>();
    int n = vars.length;
    for (int i = 0; i < n; i++)
    {
      Expression exp = new IdentifierExp(vals[i].ident);
      exp.setType(vals[i].type);
      binders.add(new Binder(vars[i].ident, exp));
    }
    Expression result = new LetParExp(binders, body);
    result.setType(body.getType());
    result.setPosition(body.getPosition());
    return result;
  }
  
  /****************************************************************************
   * Create bound expression.
   * @param vars the variables to be bound.
   * @param exp the (tuple) expression that determines the variable values.
   * @param body the body of the expression.
   * @return the bound expression.
   ***************************************************************************/
  private static Expression bindTuple(ValueSymbol[] vars, Expression exp, Expression body)
  {
    int vn = vars.length;
    List<Binder> binders = new ArrayList<Binder>();
    Identifier rident;
    if (vn == 1)
      rident = vars[0].ident;
    else
    {
      rident = new Identifier(TypeChecker.resultName);
      TupleType etype = (TupleType)exp.getType();
      rident.setSymbol(new ValueSymbol(rident, etype, ValueSymbol.Kind.value));
      List<Binder> binders0 = new ArrayList<Binder>();
      for (int i = 0; i < vn; i++)
      {
        Expression iexp = new IdentifierExp(rident);
        iexp.setType(etype);
        Expression sexp = new TupleSelectionExp(iexp, new Decimal(Integer.toString(i+1)));
        sexp.setType(etype.types[i]);
        binders0.add(new Binder(vars[i].ident, sexp));
      }
      Expression body0 = new LetParExp(binders0, body);
      body0.setType(body.getType());
      body = body0;
    }
    binders.add(new Binder(rident, exp));
    Expression result = new LetParExp(binders, body);
    result.setType(body.getType());
    return result;
  }
  
  /****************************************************************************
   * Create bound expression with binding for component from tuple expression.
   * @param csymbol the component.
   * @param exp the tuple of the values for the component variables.
   * @param body the body of the expression.
   * @return the bound expression.
   ***************************************************************************/
  private static Expression bindComponent(ComponentSymbol csymbol, Expression exp,
    Expression body)
  {
    ValueSymbol csymbol0 = componentVariable(csymbol, false, false);
    List<Binder> binders = new ArrayList<Binder>();
    binders.add(new Binder(csymbol0.ident, exp));
    List<ExpressionIdentifier> eids = new ArrayList<ExpressionIdentifier>();
    int i = 1;
    for (ValueSymbol var : csymbol.getVars())
    {
      Identifier ident = var.ident;
      Expression cexp = new IdentifierExp(csymbol0.ident);
      cexp.setType(csymbol0.type);
      Expression sexp = new TupleSelectionExp(cexp, new Decimal(Integer.toString(i)));
      ValueSymbol vsymbol = (ValueSymbol)ident.getSymbol();
      sexp.setType(vsymbol.type);
      eids.add(new ExpressionIdentifier(ident, sexp));
      i++;
    }
    Expression rexp = i == 1 ? new UnitExp() : new RecordExp(eids);
    Type rtype = i == 1 ? new UnitType() : 
      new RecordType(Arrays.asList(csymbol.getVars()), true);
    rtype.setCanonical();
    rexp.setType(rtype);
    binders.add(new Binder(csymbol0.ident, rexp));
    Expression result = new LetExp(binders, body);
    result.setType(body.getType());
    result.setPosition(body.getPosition());
    return result;
  }

  /****************************************************************************
   * Bind "old" system variables in expression by the current ones.
   * @param system the system.
   * @param exp the expression.
   * @return the bound expression.
   ***************************************************************************/
  private static Expression bindOldVars(SystemSymbol system, Expression exp)
  {
    Set<String> free = FreeVariables.compute(exp);
    ValueSymbol[] vars = system.getVars();
    ValueSymbol[] ovars = system.getOldVars();
    int n = vars.length;
    List<Binder> binders = new ArrayList<Binder>();
    for (int i = 0; i < n; i++)
    {
      Identifier oident = ovars[i].ident;
      if (free.contains(oident.string))
      {
        Identifier ident = vars[i].ident;
        ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
        Expression iexp = new IdentifierExp(ident);
        iexp.setType(symbol.type);
        binders.add(new Binder(oident, iexp));
      }
    }
    if (binders.isEmpty()) return exp;
    Expression result = new LetParExp(binders, exp);
    result.setType(exp.getType());
    return result;
  }
  
  /****************************************************************************
   * Close formula with respect to component state.
   * @param csymbol the symbol of the component in which the expression occurs.
   * @param newv true if the "new" versions of the variables are to be bound.
   * @param exp the formula.
   * @return the formula referring to the system state.
   ***************************************************************************/
  private static Expression closeComponent(ComponentSymbol csymbol, 
    boolean newv, Expression exp)
  {
    // the original position of the expression
    SourcePosition pos = exp.getPosition();
    
    // expression referring to component, possibly using the "this" index
    ValueSymbol componentIndex = csymbol.instances == 0 ? null : 
      componentIndex(csymbol, TypeChecker.thisName); 
    Expression cexp = componentExp(csymbol, componentIndex, false, newv);
    
    // bind component variables
    List<Binder> vbinders = new ArrayList<Binder>();
    ValueSymbol[] vars0 = csymbol.getVars();
    ValueSymbol[] vars = newv ? csymbol.getNewVars() : csymbol.getVars();
    int n = vars.length;
    for (int i = 0; i < n; i++)
    {
      Identifier ident0 = vars0[i].ident;
      ValueSymbol vsymbol = (ValueSymbol)ident0.getSymbol();
      Expression sexp = new RecordSelectionExp(cexp, ident0);
      sexp.setType(vsymbol.type);
      Binder vbinder = new Binder(vars[i].ident, sexp);
      vbinders.add(vbinder);
    }
    Type etype = exp.getType();
    exp = new LetParExp(vbinders, exp);
    exp.setType(etype);
    
    // the resulting expression with position unchanged
    exp.setPosition(pos);
    return exp;
  }
  
  /***************************************************************************
   * Close expression with respect to all components of system.
   * @param system the system.
   * @param exp the expression.
   * @return the expression closed with respect to component identifiers.
   *************************************************************************/
  private static Expression closeSystem(SystemSymbol system, Expression exp)
  {
    ComponentSymbol[] components = system.getComponents();
    if (components == null || components.length == 0) return exp;
    List<Binder> binders = new ArrayList<Binder>();
    for (ComponentSymbol csymbol : components)
    {
      ValueSymbol csymbol0 = componentVariable(csymbol, false, false);
      Expression cref0 = new IdentifierExp(csymbol0.ident);
      cref0.setType(csymbol0.type);
      binders.add(new Binder(csymbol.ident, cref0));
    }
    Expression exp0 = new LetParExp(binders, exp);
    exp0.setType(exp.getType());
    return exp0;
  }
  
  /***************************************************************************
   * Close list of conditions with respect to component.
   * @param csymbol the component.
   * @param conditions the conditions.
   * @return the closed versions of the conditions.
   **************************************************************************/
  private List<Condition> closeComponent(ComponentSymbol csymbol,
    List<Condition> conditions)
  {
    List<Condition> conditions0 = new ArrayList<Condition>();
    for (Condition condition : conditions)
    {
      Expression exp = closeComponent(csymbol, false, condition.exp);
      conditions0.add(condition.clone(exp));
    }
    return conditions0;
  }
  
  /***************************************************************************
   * Generate identifier for index of component instance.
   * @param csymbol the component.
   * @param the name of the identifier
   * @return the identifier symbol.
   **************************************************************************/
  private static ValueSymbol componentIndex(ComponentSymbol csymbol, String name)
  {
    Identifier ident = new Identifier(name);
    int N = csymbol.instances;
    Type type = natType(N-1);
    ValueSymbol symbol = new ValueSymbol(ident, type, ValueSymbol.Kind.value);
    ident.setSymbol(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * Return expression referring to component state.
   * @param csymbol the component.
   * @param index the index of the component instance (null, if none)
   * @param original true if the original name without underscore shall be used.
   * @param newv true if the "new" component state shall be referred.
   * is set to the identifier denoting the instance of the component.
   ***************************************************************************/
  private static Expression componentExp(ComponentSymbol csymbol, 
    ValueSymbol index, boolean original, boolean newv)
  {
    Expression iexp = null;
    if (index != null)
    {
      iexp = new IdentifierExp(index.ident);
      iexp.setType(index.type);
    }
    return componentExpCore(csymbol, iexp, original, newv);
  }
  
  /****************************************************************************
   * Return expression referring to component state.
   * @param csymbol the component.
   * @param index the index of the component instance (null, if none)
   * @param original true if the original name without underscore shall be used.
   * @param newv true if the "new" component state shall be referred.
   * is set to the identifier denoting the instance of the component.
   ***************************************************************************/
  private static Expression componentExpCore(ComponentSymbol csymbol, 
    Expression iexp, boolean original, boolean newv)
  {
    ValueSymbol vsymbol = componentVariable(csymbol, original, newv);
    Expression cexp = new IdentifierExp(vsymbol.ident);
    cexp.setType(vsymbol.type);
    if (iexp != null) 
    {
      cexp = new MapSelectionExp(cexp, iexp);
      MapType ctype0 = (MapType)vsymbol.type;
      cexp.setType(ctype0.type2);
    }
    return cexp;
  }
  
  /****************************************************************************
   * Return expression with component updated by its state variables.
   * @param csymbol the component
   * @param exp the expression
   * @return expression "let component = component with .var = var, ... in exp"
   ***************************************************************************/
  private static Expression updateComponent(ComponentSymbol csymbol, Expression exp)
  {
    ValueSymbol index = csymbol.instances == 0 ? null : 
      componentIndex(csymbol, TypeChecker.thisName);
    Expression update = componentExp(csymbol, index, false, false);
    Type type = update.getType();
    for (ValueSymbol var : csymbol.getVars())
    {
      Identifier ident = var.ident;
      Expression iexp = new IdentifierExp(ident);
      ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
      iexp.setType(symbol.type);
      update = new RecordUpdateExp(update, ident, iexp);
      update.setType(type);
    }
    if (csymbol.instances != 0)
    {
      Expression cexp = componentExp(csymbol, null, false, false);
      Expression iexp = new IdentifierExp(index.ident);
      iexp.setType(index.type);
      update = new MapUpdateExp(cexp, iexp, update);
      update.setType(cexp.getType());
    }
    List<Binder> binders = new ArrayList<Binder>();
    ValueSymbol csymbol0 = componentVariable(csymbol, false, false);
    binders.add(new Binder(csymbol0.ident, update));
    Expression result = new LetExp(binders, exp);
    result.setType(exp.getType());
    result.setPosition(exp.getPosition());
    return result;
  }
  
  /***************************************************************************
   * Get formula in context where first message from action buffer was removed.
   * @param asymbol the action.
   * @param exp the formula.
   * @return the formula embedded in the appropriate context.
   ***************************************************************************/
  private Expression removeMessage(ActionSymbol asymbol, Expression exp)
  {
    // C[this]
    ComponentSymbol csymbol = asymbol.getComponent();
    ValueSymbol cindex = csymbol.instances == 0 ? null : 
      componentIndex(csymbol, TypeChecker.thisName); 
    Expression cexp = componentExp(csymbol, cindex, false, false);
    Type ctype = cexp.getType();
    
    // action buffer and number of messages in buffer
    ValueSymbol buffer = TypeChecker.actionBuffer(asymbol);
    ValueSymbol number = TypeChecker.actionNumber(asymbol);
    MapType btype = (MapType)buffer.type;
    IntType ntype = (IntType)number.type;
    
    // the symbol for the auxiliary action buffer
    Identifier aident0 = new Identifier("_" + buffer.ident);
    ValueSymbol asymbol0 = new ValueSymbol(aident0, buffer.type, ValueSymbol.Kind.value);
    aident0.setSymbol(asymbol0);
    
    // i, i+1
    Identifier index = new Identifier("_i");
    ValueSymbol isymbol = new ValueSymbol(index, ntype, ValueSymbol.Kind.value);
    index.setSymbol(isymbol);
    Expression iexp = new IdentifierExp(index);
    iexp.setType(isymbol.type);
    Expression one = new NumberLiteralExp(new Decimal("1"));
    one.setType(new IntType(1,1));
    Expression iexp1 = new PlusExp(iexp, one);
    iexp1.setType(new IntType(ntype.ivalue1+1, ntype.ivalue2+1));
    
    // forall i with i < C[this].action_number-1. _action[i] = C[this].action[i+1}
    Expression aexp0 = new IdentifierExp(aident0);
    aexp0.setType(btype);
    aexp0 = new MapSelectionExp(aexp0, iexp);
    aexp0.setType(btype.type2);
    Expression aexp1 = new RecordSelectionExp(cexp, asymbol.ident);
    aexp1.setType(btype);
    aexp1 = new MapSelectionExp(aexp1, iexp1);
    aexp1.setType(btype.type2);
    aexp0 = new EqualsExp(aexp0, aexp1);
    aexp0.setType(new BoolType());
    Expression aexp2 = new RecordSelectionExp(cexp, number.ident);
    aexp2.setType(ntype);
    aexp2 = new MinusExp(aexp2, one);
    aexp2.setType(new IntType(ntype.ivalue1 == 0 ? 0 : ntype.ivalue1-1, 
        ntype.ivalue2 == 0 ? 0 : ntype.ivalue2-1));
    aexp2 = new LessExp(iexp, aexp2);
    aexp2.setType(new BoolType());
    List<QuantifiedVariableCore> qvc0 = new ArrayList<QuantifiedVariableCore>();
    qvc0.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(index, ntype, aexp2));
    aexp0 = new ForallExp(new QuantifiedVariable(qvc0, null), aexp0);
    aexp0.setType(new BoolType());
    
    // C[this] with .action_number = C[this].action_number-1 with .action = _action
    Expression nexp = new RecordSelectionExp(cexp, number.ident);
    nexp.setType(ntype);
    Expression nexp1 = new MinusExp(nexp, one);
    nexp1.setType(new IntType(ntype.ivalue1 == 0 ? 0 : ntype.ivalue1-1, 
        ntype.ivalue2 == 0 ? 0 : ntype.ivalue2-1));
    Expression cexp0 = new RecordUpdateExp(cexp, number.ident, nexp1);
    cexp0.setType(ctype);
    Expression aexp = new IdentifierExp(asymbol0.ident);
    aexp.setType(asymbol0.type);
    cexp0 = new RecordUpdateExp(cexp0, buffer.ident, aexp);
    cexp0.setType(ctype);
    
    // C with [this] = ...
    if (csymbol.instances > 0)
    {
      ValueSymbol cvar = componentVariable(csymbol, false, false);
      IdentifierExp cvar0 = new IdentifierExp(cvar.ident);
      cvar0.setType(cvar.type);
      IdentifierExp iexp0 = new IdentifierExp(cindex.ident);
      iexp0.setType(cindex.type);
      cexp0 = new MapUpdateExp(cvar0, iexp0, cexp0);
      Type ctype0 = componentType(csymbol);
      cexp0.setType(ctype0);
    }
    
    // forall _action with (forall i with ...) . let C = C with [this] = ... in ...
    ValueSymbol vsymbol = componentVariable(csymbol, false, false);
    List<Binder> binders = new ArrayList<Binder>();
    binders.add(new Binder(vsymbol.ident, cexp0));
    Expression exp0 = new LetExp(binders, exp);
    exp0.setType(exp.getType());
    List<QuantifiedVariableCore> qvc = new ArrayList<QuantifiedVariableCore>();
    qvc.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(asymbol0.ident, asymbol0.type, aexp0));
    exp0 = new ForallExp(new QuantifiedVariable(qvc, null), exp0);
    exp0.setType(exp.getType());
    return exp0;
  }
  
  // --------------------------------------------------------------------------
  //
  // send command
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Substitute send command by assignments to state.
   * @param command the send command.
   * @return the list of assignments.
   ***************************************************************************/
  public List<Command> sendCommand(SendCommand command)
  {
    // C[i].send(m)
    // -> C[i].send[C[i].send_number] := m; 
    //    C[i].send_number := C[i].send_number+1;
    
    // the message buffer and the number of elements in that buffer
    ActionSymbol asymbol = (ActionSymbol)command.ident2.getSymbol();
    ValueSymbol buffer = TypeChecker.actionBuffer(asymbol);
    ValueSymbol number = TypeChecker.actionNumber(asymbol);
    
    // the message m
    int n = command.exps.length;
    List<ExpressionIdentifier> eidents = new ArrayList<ExpressionIdentifier>();
    for (int i = 0; i < n; i++)
    {
      ValueSymbol param = asymbol.params[i];
      eidents.add(new ExpressionIdentifier(param.ident, command.exps[i]));
    }
    Expression message = n == 0 ? new UnitExp() :
      new RecordExp(eidents);
    Type rtype = n == 0 ? new UnitType() :
      new RecordType(Arrays.asList(asymbol.params), true);
    rtype.setCanonical();
    message.setType(rtype);
    
    // the component identifier C
    ComponentSymbol csymbol = asymbol.getComponent();
    ValueSymbol vsymbol = componentVariable(csymbol, false, false);
    
    // the number expression C[i].send_number
    Expression cexp = componentExpCore(csymbol, command.exp, false, false);
    Expression nexp = new RecordSelectionExp(cexp, number.ident);
    nexp.setType(number.type);
    
    // the buffer selectors [i].send[C[i].send_number]
    List<Selector> bsels = new ArrayList<Selector>();
    if (command.exp != null) bsels.add(new Selector.MapSelector(command.exp));
    bsels.add(new Selector.RecordSelector(buffer.ident));
    bsels.add(new Selector.MapSelector(nexp));

    // the command C[i].send[C[i].send_number] := m; 
    Command command1 = new AssignmentCommand(vsymbol.ident, bsels, message);
    
    // the number selectors [i].send_number
    List<Selector> nsels = new ArrayList<Selector>();
    if (command.exp != null) nsels.add(new Selector.MapSelector(command.exp));
    nsels.add(new Selector.RecordSelector(number.ident));
    
    // the expression C[i].send_number+1
    Expression one = new NumberLiteralExp(new Decimal("1"));
    one.setType(new IntType(1,1));
    Expression nexp1 = new PlusExp(nexp, one);
    IntType ntype = (IntType)number.type;
    nexp1.setType(new IntType(ntype.ivalue1+1, ntype.ivalue2+1));

    // the command C[i].send_number := C[i].send_number+1;
    Command command2 = new AssignmentCommand(vsymbol.ident, nsels, nexp1);
    
    // the complete translation
    List<Command> result = new ArrayList<Command>();
    result.add(command1);
    result.add(command2);
    return result;
  }
  
  /****************************************************************************
   * Get expression that action buffer of send command is not full.
   * @param command the send command.
   * @return the expression.
   ***************************************************************************/
  public Expression notFullExp(SendCommand command)
  {
    ActionSymbol asymbol = (ActionSymbol)command.ident2.getSymbol();
    ComponentSymbol csymbol = asymbol.getComponent();
    Expression cexp = componentExpCore(csymbol, command.exp, false, false);
    ValueSymbol number = TypeChecker.actionNumber(asymbol);
    Expression nexp = new RecordSelectionExp(cexp, number.ident);
    nexp.setType(number.type);
    Expression bsize = new NumberLiteralExp(new Decimal(Integer.toString(asymbol.bsize)));
    Type btype = new IntType(asymbol.bsize, asymbol.bsize);
    btype.setCanonical();
    bsize.setType(btype);
    Expression lessExp = new LessExp(nexp, bsize);
    lessExp.setType(new BoolType());
    return lessExp;
  }
}

// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------